/*******************************************************************************

    File:     cdh_maps2.c
    Author:   Rick Cooper
    Created:  16/05/12
    Edited:   16/05/12
    Contents: A front-end to produce PDF (via ciaro) of maps for the CDH simulations

    Public procedures:
        int main(int argc, char **argv)

*******************************************************************************/

#include "src_model.h"
#include "lib_maths.h"
#include <stdlib.h>
#include <time.h>
#include <gtk/gtk.h>

#include <string.h>
#include <cairo/cairo-pdf.h>
#include "lib_gtkx.h"
#include "lib_cairox.h"

/******************************************************************************/

typedef enum dv {
    DV_PSP0, DV_PSP1, DV_PSP2, DV_FIT, DV_MAX
} DV;

#define SCALE 1.25

#define IV_LR_MAX 31
#define IV_TAU_MAX 11

#define W_PIXELS IV_LR_MAX
#define H_PIXELS IV_TAU_MAX

#define PID_MAX 101

char *graph_label[DV_MAX] = {
    "NC versus C",
    "NC versus S",
    "S versus C",
    "General Fit",
};

int counts[IV_LR_MAX][IV_TAU_MAX];
double simulated_values[PID_MAX][IV_LR_MAX][IV_TAU_MAX][DV_MAX];
double simulated_mean[IV_LR_MAX][IV_TAU_MAX][DV_MAX];

double iv_lr_values[IV_LR_MAX];
double iv_tau_values[IV_TAU_MAX];
int lr_index[IV_LR_MAX];
int tau_index[IV_TAU_MAX];

double data[W_PIXELS][H_PIXELS];

/******************************************************************************/


/* We use this to indicate that an IV's value has not been set */
#define MAX_NEG -999999

/*----------------------------------------------------------------------------*/

static double dabs(double x)
{
    return((x > 0) ? x : -x);
}

static int map_lr(double lr)
{
    int i;

    for (i = 0; i < IV_LR_MAX; i++) {
        if (iv_lr_values[i] == MAX_NEG) {
	    iv_lr_values[i] = lr;
	    return(i);
	}
	else if (dabs(iv_lr_values[i] - lr) < 0.00001) {
	    return(i);
	}
    }
    return(-1);
}

static int map_tau(int tau)
{
    int i;

    for (i = 0; i < IV_LR_MAX; i++) {
        if (iv_tau_values[i] == MAX_NEG) {
	    iv_tau_values[i] = tau;
	    return(i);
	}
	else if (dabs(iv_tau_values[i] - tau) < 0.00001) {
	    return(i);
	}
    }
    return(-1);
}

/*----------------------------------------------------------------------------*/

static void initialise_counts()
{
    int i, j;

    for (i = 0; i < IV_LR_MAX; i++) {
        for (j = 0; j < IV_TAU_MAX; j++) {
                counts[i][j] = 0;
        }
    }
    for (i = 0; i < IV_LR_MAX; i++) {
        iv_lr_values[i] = MAX_NEG;
    }
    for (j = 0; j < IV_TAU_MAX; j++) {
        iv_tau_values[j] = MAX_NEG;
    }
}

static void ignore_to_eoln(FILE *fp)
{
    int c;

    do {
        c = getc(fp);
    } while ((c > 0) && (c != '\n'));
}

static int read_data_line_psp(FILE *fp)
{
    double psp0, psp1, psp2, tau, lr, c1_strength, c2_strength;
    int pid;

    if (fscanf(fp, "%d %lf %lf %lf %lf %lf %lf %lf\n", &pid, &lr, &tau, &c1_strength, &c2_strength, &psp0, &psp1, &psp2) > 0) {
        int s = map_lr(lr);
        int i = map_tau(tau);

        if (s < 0) {
            fprintf(stdout, "Too many values of learning rate\n");
        }
        else if (i < 0) {
            fprintf(stdout, "Too many values of tau\n");
        }
        else if (counts[s][i] < PID_MAX) {
            simulated_values[counts[s][i]][s][i][DV_PSP0] = psp0;
            simulated_values[counts[s][i]][s][i][DV_PSP1] = psp1;
            simulated_values[counts[s][i]][s][i][DV_PSP2] = psp2;
            counts[s][i]++;
        }
        else {
            fprintf(stdout, "Too many participants (lr: %f; tau: %f)\n", lr, tau);
        }
        return(TRUE);
    }
    else {
        return(FALSE);
    }
}

int data_read_psp_scores(char *filename)
{
    int count;
    FILE *fp;

    initialise_counts();
    if ((fp = fopen(filename, "r")) == NULL) {
        fprintf(stderr, "Failed to open %s\n", filename);
        return(0);
    }
    else {
        ignore_to_eoln(fp);
        count = 0;
        while (read_data_line_psp(fp)) {
            count++;
        }
        fprintf(stdout, "Read %d records from %s\n", count, filename);
        return(1);
    }
}

/******************************************************************************/

static void set_psp_colour(cairo_t *cr, int bw, double data)
{
    double d = ((data + 1.0) / 2.0);

    if (bw) {
        cairo_set_source_rgb(cr, d, d, d);
    }
    else {
        cairox_select_colour_scale(cr, d);
    }
}
 
/******************************************************************************/

double calculate_fit(int i, int j)
{
    double f;
    // Ideally we want: DV_PSP0 = -1; DV_PSP1 = -1; DV_PSP2 =  0

    // We calculate fit as the geometric mean of the separate values, scaled to the zero to 1 range: 

    f = (1.0 - simulated_mean[i][j][DV_PSP0]) * 0.5 * 
        (1.0 - simulated_mean[i][j][DV_PSP1]) * 0.5 * 
        (1.0 - fabs(simulated_mean[i][j][DV_PSP2]));

    f = exp(log(f) / 3);

    // Alternatively we could calculate fit as the minimum of the separate values:
//    f = (1.0 - simulated_mean[i][j][DV_PSP0]) * 0.5;
//    f = MIN(f, (1.0 - simulated_mean[i][j][DV_PSP1]) * 0.5);
//    f = MIN(f, (1.0 - fabs(simulated_mean[i][j][DV_PSP2])));

    // f ranges from 0 (no fit) to 1 (perfect fit) - rescale to [-1,+1]

    return(1.0 - (2.0 * f));
}

static void collate_simulated_dvs()
{
    int i, j, m, n;
    
    for (i = 0; i < IV_LR_MAX; i++) {
        for (j = 0; j < IV_TAU_MAX; j++) {
            for (m = 0; m < (DV_MAX-1); m++) {
                if (counts[i][j] > 0) {
                    double sum = 0;

                    for (n = 0; n < counts[i][j]; n++) {
                        sum += simulated_values[n][i][j][m];
                    }
                    simulated_mean[i][j][m] = sum / (double) counts[i][j];
                }
                else {
                    simulated_mean[i][j][m] = 0.0;
                }
            }
            simulated_mean[i][j][DV_FIT] = calculate_fit(i, j);
        }
    }
}

static void generate_indicies()
{
    // Populate arrays with rank-ordered indices for the IVs:

    int i, j, k;

    for (i = 0; i < IV_LR_MAX; i++) {
        lr_index[i] = i;
    }
    for (i = 0; i < IV_LR_MAX; i++) {
        for (j = i+1; j < IV_LR_MAX; j++) {
            if (iv_lr_values[lr_index[i]] > iv_lr_values[lr_index[j]]) {
                k = lr_index[i]; lr_index[i] = lr_index[j]; lr_index[j] = k;
	    }
        }
    }

    for (i = 0; i < IV_TAU_MAX; i++) {
        tau_index[i] = i;
    }
    for (i = 0; i < IV_TAU_MAX; i++) {
        for (j = i+1; j < IV_TAU_MAX; j++) {
            if (iv_tau_values[tau_index[i]] > iv_tau_values[tau_index[j]]) {
                k = tau_index[i]; tau_index[i] = tau_index[j]; tau_index[j] = k;
	    }
        }
    }
}

static void draw_grid_psp(cairo_t *cr, PangoLayout *layout, char *axis_label, int x0, int y0)
{
    CairoxTextParameters p;
    char buffer[32];
    int i;

    cairo_set_line_width(cr, 1.0 / SCALE);

    /* Paint in white to stop bleeding around the top/right of the map */
    cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
    cairox_paint_line(cr, 11.0, x0 - 6, y0 - 6, x0 + 106, y0 - 6);
    cairox_paint_line(cr, 11.0, x0 - 6, y0 + 106, x0 + 106, y0 + 106);
    cairox_paint_line(cr, 11.0, x0 - 6, y0 - 6, x0 - 6, y0 + 106);
    cairox_paint_line(cr, 11.0, x0 + 106, y0 - 6, x0 + 106, y0 + 106);

    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);

    /* Vertical grid */
    for (i = 0; i < 7; i++) {
        cairox_paint_line(cr, 0.5, x0 + i * 100/6.0, y0, x0 + i * 100/6.0, y0 + 100);
    }

    /* Horizontal grid */
    for (i = 0; i < 11; i++) {
        cairox_paint_line(cr, 0.5, x0, y0 + i * 10, x0 + 100, y0 + i * 10);
    }

    /* X (horizontal) axis labels: */

    pangox_layout_set_font_size(layout, 8 / SCALE);
    for (i = 0; i < 7; i++) {
        sprintf(buffer, "%5.3f", (iv_lr_values[lr_index[0]] + i * (iv_lr_values[lr_index[IV_LR_MAX-1]] - iv_lr_values[lr_index[0]]) / 6.0));
        cairox_text_parameters_set(&p, x0+i*100/6.0-7, y0+106, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 45.0);
        cairox_paint_pango_text(cr, &p, layout, buffer);
    }

    pangox_layout_set_font_size(layout, 10 / SCALE);
    cairox_text_parameters_set(&p, x0+50, y0+118, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "Learning Rate");

    /* Y (vertical) axis labels: */

    pangox_layout_set_font_size(layout, 8 / SCALE);

    for (i = 0; i < 11; i++) {
        sprintf(buffer, "%3d", (int) (iv_tau_values[tau_index[0]] + i * (iv_tau_values[tau_index[IV_TAU_MAX-1]] - iv_tau_values[tau_index[0]]) / 10.0));
        cairox_text_parameters_set(&p, x0-3, y0+100-i*10, PANGOX_XALIGN_RIGHT, PANGOX_YALIGN_CENTER, 0.0);
        cairox_paint_pango_text(cr, &p, layout, buffer);
    }

    pangox_layout_set_font_size(layout, 10 / SCALE);
    cairox_text_parameters_set(&p, x0-23, y0+50, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_TOP, 90.0);
    cairox_paint_pango_text(cr, &p, layout, axis_label);
}

static void draw_title(cairo_t *cr, PangoLayout *layout, double x0, double y0, char *title)
{
    CairoxTextParameters p;

    cairo_set_line_width(cr, 1.0 / SCALE);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
    cairo_rectangle(cr, x0, y0, 100, 100);
    cairo_stroke(cr);

    pangox_layout_set_font_size(layout, 10 / SCALE);
    cairox_text_parameters_set(&p, x0+50, y0-4, PANGOX_XALIGN_CENTER, PANGOX_YALIGN_BOTTOM, 0.0);
    cairox_paint_pango_text(cr, &p, layout, title);
}

/******************************************************************************/

static void load_graph_data(int dv)
{
    int COMPRESSION_X = (W_PIXELS-1) / (IV_LR_MAX-1);
    int COMPRESSION_Y = (H_PIXELS-1) / (IV_TAU_MAX-1);
    int i, j;

    for (i = 0; i < W_PIXELS; i++) {
        for (j = 0; j < H_PIXELS; j++) {
            int xw1, xw2, yw1, yw2;
            int ei1 = 0, ei2 = 0, ii1 = 0, ii2 = 0;

            /* Weights: */
            xw1 = COMPRESSION_X * (1 + (i / COMPRESSION_X)) - i;
            xw2 = i - COMPRESSION_X * (i / COMPRESSION_X);

            yw1 = COMPRESSION_Y * (1 + (j / COMPRESSION_Y)) - j;
            yw2 = j - COMPRESSION_Y * (j / COMPRESSION_Y);

            /* Indices: */
            if (xw1 != 0) {
                ei1 = lr_index[i / COMPRESSION_X];
            }
            if (xw2 != 0) {
                ei2 = lr_index[1 + (i / COMPRESSION_X)];
            }
            if (yw1 != 0) {
                ii1 = tau_index[j / COMPRESSION_Y];
            }
            if (yw2 != 0) {
                ii2 = tau_index[1 + (j / COMPRESSION_Y)];
            }

            data[i][j] = (xw1 * yw1 * simulated_mean[ei1][ii1][dv] + 
                          xw1 * yw2 * simulated_mean[ei1][ii2][dv] + 
                          xw2 * yw1 * simulated_mean[ei2][ii1][dv] + 
                          xw2 * yw2 * simulated_mean[ei2][ii2][dv]) / 
                        ((xw1 + xw2) * (yw1 + yw2));

        }
    }
}

static void draw_graph(cairo_t *cr, int bw, int x0, int y0)
{
    double voxel_W = 100 / (double) (W_PIXELS - 1);
    double voxel_H = 100 / (double) (H_PIXELS - 1);
    int i, j;

    for (i = 0; i < W_PIXELS; i++) {
        for (j = 0; j < H_PIXELS; j++) {
            set_psp_colour(cr, bw, data[i][j]);
            cairo_rectangle(cr, x0 + (i-0.5) * voxel_W, y0 + 100 - (j+0.5) * voxel_H, voxel_W, voxel_H);
            cairo_fill(cr);
	}
    }
}

/******************************************************************************/

static void draw_surface_psp(cairo_t *cr, PangoLayout *layout, int bw, double x0, double y0, DV var, char *graph_label, char *axis_label)
{ 
    load_graph_data(var);
    draw_graph(cr, bw, x0, y0);
    draw_grid_psp(cr, layout, axis_label, x0, y0);
    draw_title(cr, layout, x0, y0, graph_label);
}

static void draw_legend(cairo_t *cr, PangoLayout *layout, int bw, double x0, double y0)
{
    CairoxTextParameters p;
    double y = y0;

    pangox_layout_set_font_size(layout, 10 / SCALE);

    cairo_rectangle(cr, x0, y, 5, 5);
    set_psp_colour(cr, bw, -1.0);
    cairo_fill_preserve(cr);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
    cairo_stroke(cr);

    cairox_text_parameters_set(&p, x0+10, y+6, PANGOX_XALIGN_LEFT, PANGOX_YALIGN_BOTTOM, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "X > Y");

    y += 15;

    cairo_rectangle(cr, x0, y, 5, 5);
    set_psp_colour(cr, bw, 0.0);
    cairo_fill_preserve(cr);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
    cairo_stroke(cr);

    cairox_text_parameters_set(&p, x0+10, y+6, PANGOX_XALIGN_LEFT, PANGOX_YALIGN_BOTTOM, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "X = Y");

    y += 15;

    cairo_rectangle(cr, x0, y, 5, 5);
    set_psp_colour(cr, bw, 1.0);
    cairo_fill_preserve(cr);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
    cairo_stroke(cr);

    cairox_text_parameters_set(&p, x0+10, y+6, PANGOX_XALIGN_LEFT, PANGOX_YALIGN_BOTTOM, 0.0);
    cairox_paint_pango_text(cr, &p, layout, "X < Y");
}

/******************************************************************************/

static void lr_maps_draw(GlobalVariables *gv, cairo_t *cr, int bw, char *title, char *axis_label)
{
    CairoxTextParameters p;
    PangoLayout *layout;
    char buffer[32];

    /* Clear the background: */
    cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
    cairo_paint(cr);

    cairo_save(cr);
    cairo_scale(cr, SCALE, SCALE);
    layout = pango_cairo_create_layout(cr);
    pangox_layout_set_font_face(layout, PS_FONT_SANS);

    draw_surface_psp(cr, layout, bw,  30,  30, DV_PSP0, graph_label[DV_PSP0], axis_label);
    draw_surface_psp(cr, layout, bw, 180,  30, DV_PSP1, graph_label[DV_PSP1], axis_label);
    draw_surface_psp(cr, layout, bw, 330,  30, DV_PSP2, graph_label[DV_PSP2], axis_label);

    draw_legend(cr, layout, bw, 460, 30);

    pangox_layout_set_font_size(layout, 12 / SCALE);
    sprintf(buffer, "Learning rule: %s", title);
    cairox_text_parameters_set(&p, 10, 0, PANGOX_XALIGN_LEFT, PANGOX_YALIGN_TOP, 0.0);
    cairox_paint_pango_text(cr, &p, layout, buffer);

    cairo_restore(cr);
    g_object_unref(layout);
}

static void draw_fit_legend(cairo_t *cr, PangoLayout *layout, int bw, double x0, double y0)
{
    CairoxTextParameters p;
    char buffer[8];
    int i;

    for (i = 0; i < 100; i++) {
        set_psp_colour(cr, bw, i * 0.02 - 1.0);
        cairo_rectangle(cr, x0-3, y0+i-0.1, 6, 1.2);
        cairo_fill(cr);
    }

    pangox_layout_set_font_size(layout, 8 / SCALE);
    cairo_set_source_rgb(cr, 0.0, 0.0, 0.0);
    cairo_rectangle(cr, x0-3, y0, 6, 100);
    cairo_stroke(cr);

    for (i = 0; i < 6; i++) {
        cairox_paint_line(cr, 1.0, x0-2.9, y0+i*20.0, x0+5.0, y0+i*20.0);
        sprintf(buffer, "%4.2f", 1.0 - (i * 0.2));
        cairox_text_parameters_set(&p, x0+6, y0+i*20, PANGOX_XALIGN_LEFT, PANGOX_YALIGN_CENTER, 0.0);
        cairox_paint_pango_text(cr, &p, layout, buffer);
    }
}

/******************************************************************************/

void maps_create(GlobalVariables *gv)
{
    cairo_surface_t *surface;
    PangoLayout *layout;
    cairo_t *cr;
//    int bw = 0; // Draw in colour
    int bw = 1; // Draw in Black/White

    // Repeat for each learning rule: 
    if (data_read_psp_scores("../Results/contingency_s2_hebb_psp.txt")) {
        collate_simulated_dvs();
        generate_indicies();
        surface = cairo_pdf_surface_create("../Figures/image_map_hebb_bw.pdf", 500*SCALE, 160*SCALE);
        cr = cairo_create(surface);
        lr_maps_draw(gv, cr, bw, "Hebbian", "Asymptote");
        cairo_destroy(cr);
        cairo_surface_destroy(surface);
    }

    if (data_read_psp_scores("../Results/contingency_s2_qebb_psp.txt")) {
        collate_simulated_dvs();
        generate_indicies();
        surface = cairo_pdf_surface_create("../Figures/image_map_qebb_bw.pdf", 500*SCALE, 160*SCALE);
        cr = cairo_create(surface);
        lr_maps_draw(gv, cr, bw, "Quasi-Hebbian", "Asymptote");
        cairo_destroy(cr);
        cairo_surface_destroy(surface);
    }

    if (data_read_psp_scores("../Results/contingency_s2_rw_psp.txt")) {
        collate_simulated_dvs();
        generate_indicies();
        surface = cairo_pdf_surface_create("../Figures/image_map_rw_bw.pdf", 500*SCALE, 160*SCALE);
        cr = cairo_create(surface);
        lr_maps_draw(gv, cr, bw, "Rescorla-Wagner", "Target");
        cairo_destroy(cr);
        cairo_surface_destroy(surface);
    }

    // Now draw the composite maps

    surface = cairo_pdf_surface_create("../Figures/image_map_fits_bw.pdf", 500*SCALE, 160*SCALE);
    cr = cairo_create(surface);

    layout = pango_cairo_create_layout(cr);
    pangox_layout_set_font_face(layout, PS_FONT_SANS);

    /* Clear the background: */
    cairo_set_source_rgb(cr, 1.0, 1.0, 1.0);
    cairo_paint(cr);

    cairo_save(cr);
    cairo_scale(cr, SCALE, SCALE);

    if (data_read_psp_scores("../Results/contingency_s2_hebb_psp.txt")) {
        collate_simulated_dvs();
        generate_indicies();
        draw_surface_psp(cr, layout, bw,  30, 30, DV_FIT, "Hebbian", "Asymptote");
    }
    if (data_read_psp_scores("../Results/contingency_s2_qebb_psp.txt")) {
        collate_simulated_dvs();
        generate_indicies();
        draw_surface_psp(cr, layout, bw, 180, 30, DV_FIT, "Quasi-Hebbian", "Asymptote");
    }
    if (data_read_psp_scores("../Results/contingency_s2_rw_psp.txt")) {
        collate_simulated_dvs();
        generate_indicies();
        draw_surface_psp(cr, layout, bw, 330, 30, DV_FIT, "Rescorla-Wagner", "Target");
    }

    draw_fit_legend(cr, layout, bw, 460, 30);

    g_object_unref(layout);
    cairo_restore(cr);
    cairo_destroy(cr);
    cairo_surface_destroy(surface);

}

/******************************************************************************/
/* Section X: Main -----------------------------------------------------------*/

int main(int argc, char **argv)
{
    GlobalVariables *globals;

    random_randomise_seed();

    if ((globals = globals_create()) == NULL) {
        fprintf(stderr, "%s: Aborting - Memory allocation failed while creating global variable space\n", argv[0]);
    }
    else {
        maps_create(globals);
    }

    globals_destroy(globals);

    exit(0);
}

/******************************************************************************/
