#include "src_model.h"
#include "cdh_task.h"
#include "lib_maths.h"
#include "lib_gtkx.h"
#include "lib_cairox.h"

double subject_means[NUM_TR][NUM_CONDITIONS][MAX_SUBJECTS];

/******************************************************************************/

void print_subject_stats(FILE *fp, SubjectResults *results, int blocks)
{
    int i, j;

    for (i = 0; i < blocks; i++) {
        for (j = 0; j < NUM_CONDITIONS; j++) {
            fprintf(stdout, "Response[Block=%d][Condition=%d]: %d (RT = %f)\n", i, j, results->response[i][j], results->rt[i][j]);
        }
    }
}

/******************************************************************************/

double exp1_get_condition_rt_mean(GlobalVariables *gv, TrainingType tt, int condition)
{
    double sum;
    int i;

    sum = 0.0;
    for (i = 0; i < gv->subject_count1; i++) {
        sum += subject_means[(int) tt][condition][i];
    }
    return((gv->subject_count1 > 0) ? (sum / (double) gv->subject_count1) : 0.0);
}

double exp1_get_condition_rt_se(GlobalVariables *gv, TrainingType tt, int condition)
{
    double sum, ssq;
    int i;

    sum = 0.0; ssq = 0.0;
    for (i = 0; i < gv->subject_count1; i++) {
        sum += subject_means[(int) tt][condition][i];
        ssq += subject_means[(int) tt][condition][i] * subject_means[(int) tt][condition][i];
    }
    return((gv->subject_count1 > 1) ? sqrt(((ssq * gv->subject_count1) - (sum * sum)) / (double) (gv->subject_count1 * (gv->subject_count1 - 1))) / (sqrt(gv->subject_count1)) : 0.0);
}

double exp1_get_condition_difference_mean(GlobalVariables *gv, TrainingType tt)
{
    double sum;
    int i;

    sum = 0.0;
    for (i = 0; i < gv->subject_count1; i++) {
        sum += (subject_means[(int) tt][1][i] - subject_means[(int) tt][0][i]);
    }
    return((gv->subject_count1 > 0) ? (sum / (double) gv->subject_count1) : 0.0);
}

double exp1_get_condition_difference_se(GlobalVariables *gv, TrainingType tt)
{
    double sum, ssq, diff;
    int i;

    sum = 0.0; ssq = 0.0;
    for (i = 0; i < gv->subject_count1; i++) {
        diff = (subject_means[(int) tt][1][i] - subject_means[(int) tt][0][i]);
        sum += diff;
        ssq += diff * diff;
    }
    return((gv->subject_count1 > 1) ? sqrt(((ssq * gv->subject_count1) - (sum * sum)) / (double) (gv->subject_count1 * (gv->subject_count1 - 1))) / (sqrt(gv->subject_count1)) : 0.0);
}

double exp1_get_condition_average_mean(GlobalVariables *gv, TrainingType tt)
{
    double sum;
    int i;

    sum = 0.0;
    for (i = 0; i < gv->subject_count1; i++) {
        sum += ((subject_means[(int) tt][1][i] + subject_means[(int) tt][0][i]))/2.0;
    }
    return((gv->subject_count1 > 0) ? (sum / (double) gv->subject_count1) : 0.0);
}

double exp1_get_condition_average_se(GlobalVariables *gv, TrainingType tt)
{
    double sum, ssq, av;
    int i;

    sum = 0.0; ssq = 0.0;
    for (i = 0; i < gv->subject_count1; i++) {
        av = ((subject_means[(int) tt][1][i] + subject_means[(int) tt][0][i]))/2.0;
        sum += av;
        ssq += av * av;
    }
    return((gv->subject_count1 > 1) ? sqrt(((ssq * gv->subject_count1) - (sum * sum)) / (double) (gv->subject_count1 * (gv->subject_count1 - 1))) / (sqrt(gv->subject_count1)) : 0.0);
}

double exp1_get_condition_difference_overall_mean(GlobalVariables *gv)
{
    TrainingType tt;
    double sum;
    int i;

    sum = 0.0;
    for (i = 0; i < gv->subject_count1; i++) {
        tt = TR_CONTINGENT;
        sum += ((subject_means[(int) tt][1][i] - subject_means[(int) tt][0][i]));
        tt = TR_NONCONTINGENT;
        sum += ((subject_means[(int) tt][1][i] - subject_means[(int) tt][0][i]));
    }
    return((gv->subject_count1 > 0) ? (sum / ((double) 2.0 * gv->subject_count1)) : 0.0);
}

double exp1_get_condition_difference_overall_se(GlobalVariables *gv)
{
    TrainingType tt;
    double sum, ssq, diff;
    int i;

    sum = 0.0; ssq = 0.0;
    for (i = 0; i < gv->subject_count1; i++) {
        tt = TR_CONTINGENT;
        diff = (subject_means[(int) tt][1][i] - subject_means[(int) tt][0][i]);
        sum += diff;
        ssq += diff * diff;
        tt = TR_NONCONTINGENT;
        diff = (subject_means[(int) tt][1][i] - subject_means[(int) tt][0][i]);
        sum += diff;
        ssq += diff * diff;
    }
    return((gv->subject_count1 > 1) ? sqrt(((ssq * 2.0*gv->subject_count1) - (sum * sum)) / (double) (2.0*gv->subject_count1 * (2.0*gv->subject_count1 - 1))) / (sqrt(2.0*gv->subject_count1)) : 0.0);
}

/******************************************************************************/

double exp1_get_condition_difference_t(GlobalVariables *gv)
{
    /* Need a between-subjects t test */

//    int i;

//    fprintf(stdout, "Subject data\n");
//    for (i = 0; i < gv->subject_count1; i++) {
//        fprintf(stdout, "%7.3f %7.3f\n", subject_means[TR_CONTINGENT][1][i] - subject_means[TR_CONTINGENT][0][i], subject_means[TR_NONCONTINGENT][1][i] - subject_means[TR_NONCONTINGENT][0][i]);
//    }

    if (gv->subject_count1 > 1) {
        double mean[2], var[2], var_pooled, t;

        mean[0] = exp1_get_condition_difference_mean(gv, TR_CONTINGENT);
        mean[1] = exp1_get_condition_difference_mean(gv, TR_NONCONTINGENT);
        var[0] = exp1_get_condition_difference_se(gv, TR_CONTINGENT) * exp1_get_condition_difference_se(gv, TR_CONTINGENT) * gv->subject_count1;
        var[1] = exp1_get_condition_difference_se(gv, TR_NONCONTINGENT) * exp1_get_condition_difference_se(gv, TR_NONCONTINGENT) * gv->subject_count1;

        var_pooled = (var[0] + var[1]) / 2.0;

//        fprintf(stdout, "Means: %f %f\n", mean[0], mean[1]);
//        fprintf(stdout, "Variances: %f %f (%f)\n", var[0], var[1], var_pooled);

        t = (mean[1] - mean[0]) / sqrt(var_pooled * 2.0 / gv->subject_count1);

        return(t);
    }
    else {
        return(0);
    }
}

/******************************************************************************/

static void record_subject_means1(GlobalVariables *gv, TrainingType tt, SubjectResults results[2])
{
    double sum;
    int c, i;

    for (c = 0; c < NUM_CONDITIONS; c++) {
        if (gv->blocks1 == 0) {
            subject_means[(int) tt][c][gv->subject_count1] = 0;
        }
        else {
            sum = 0.0;
            for (i = 0; i < gv->blocks1; i++) {
                // Assumes trial isn't an error - should check results->response[i][j], 
                sum += results[0].rt[i][c];
                sum += results[1].rt[i][c];
            }
            subject_means[(int) tt][c][gv->subject_count1] = (sum / (2.0 * gv->blocks1));
        }
    }
}

/******************************************************************************/

void exp1_print_analysis(FILE *fp, GlobalVariables *gv)
{
    double means[8], ses[8];
    double var_pooled, t, diff;
    int df;

    fprintf(fp, "Learning rule: %s\n", lr_name[gv->learning_rule]);

    means[0] = exp1_get_condition_rt_mean(gv, TR_CONTINGENT, 0);
    ses[0] = exp1_get_condition_rt_se(gv, TR_CONTINGENT, 0);
    means[1] = exp1_get_condition_rt_mean(gv, TR_CONTINGENT, 1);
    ses[1] = exp1_get_condition_rt_se(gv, TR_CONTINGENT, 1);
    means[2] = exp1_get_condition_rt_mean(gv, TR_NONCONTINGENT, 0);
    ses[2] = exp1_get_condition_rt_se(gv, TR_NONCONTINGENT, 0);
    means[3] = exp1_get_condition_rt_mean(gv, TR_NONCONTINGENT, 1);
    ses[3] = exp1_get_condition_rt_se(gv, TR_NONCONTINGENT, 1);

    means[4] = exp1_get_condition_difference_mean(gv, TR_CONTINGENT);
    ses[4] = exp1_get_condition_difference_se(gv, TR_CONTINGENT);
    means[5] = exp1_get_condition_difference_mean(gv, TR_NONCONTINGENT);
    ses[5] = exp1_get_condition_difference_se(gv, TR_NONCONTINGENT);

    means[6] = exp1_get_condition_average_mean(gv, TR_CONTINGENT);
    ses[6] = exp1_get_condition_average_se(gv, TR_CONTINGENT);
    means[7] = exp1_get_condition_average_mean(gv, TR_NONCONTINGENT);
    ses[7] = exp1_get_condition_average_se(gv, TR_NONCONTINGENT);

    means[8] = exp1_get_condition_difference_overall_mean(gv);
    ses[8] = exp1_get_condition_difference_overall_se(gv);

    /* Main effect of trial type (significant in subject data): */

    df = (2 * gv->subject_count1) - 1;
    var_pooled = ses[8] * ses[8] * 2 * gv->subject_count1;
    t = means[8] / sqrt(var_pooled / (2.0 * gv->subject_count1));

    fprintf(fp, "Compatability: Diff = %4.2f units; t(%d) = %6.3f; p = %6.4f\n", means[8], df, t, prob_of_t(fabs(t), df));

    /* Main effect of group (not significant in subject data) */

    diff = means[7] - means[6];
    df = 2 * (gv->subject_count1 - 1);
    var_pooled = (ses[6] * ses[6] + ses[7] * ses[7]) * gv->subject_count1 * 0.5;
    t = (means[7] - means[6]) / sqrt(var_pooled * 2.0 / gv->subject_count1);

    fprintf(fp, "Group: Diff = %4.2f units; t(%d) = %6.3f; p = %6.4f\n", diff, df, t, prob_of_t(fabs(t), df));

    /* Work out the interaction effect (significant in subject data): */

    diff = means[5] - means[4];
    df = 2 * (gv->subject_count1 - 1);
    var_pooled = (ses[4] * ses[4] + ses[5] * ses[5]) * gv->subject_count1 * 0.5;
    t = (means[5] - means[4]) / sqrt(var_pooled * 2.0 / gv->subject_count1);

    fprintf(fp, "Interaction: Diff = %4.2f units; t(%d) = %6.3f; p = %6.4f\n", diff, df, t, prob_of_t(fabs(t), df));
}

/******************************************************************************/

void exp1_do_task(GlobalVariables *gv)
{
    SubjectResults results[2];
    int s;

    gv->subject_count1 = 0;

    for (s = 0; s < gv->subjects1; s++) {

        /* Initialise network then give contingent training */
        /* Test for 120 trials where imperative is 1 and 120 when it is 2 */
        /* Imitative stimulus is compatible on 60, incomptible on 60 of each */

        network_weights_initialise(gv);
        network_set_training_data(gv, TR_CONTINGENT);
        network_weights_train(gv, FALSE);

        /* Block 1: Always open: */
        gv->sensory_clamp_settings[0]    = 1;
        gv->imperative_clamp_settings[0] = 1;
        gv->sensory_clamp_settings[1]    = 2;
        gv->imperative_clamp_settings[1] = 1;
        network_step_block(gv, &results[0], gv->blocks1, FALSE);

        /* Block 2: Always close: */
        gv->sensory_clamp_settings[0]    = 2;
        gv->imperative_clamp_settings[0] = 2;
        gv->sensory_clamp_settings[1]    = 1;
        gv->imperative_clamp_settings[1] = 2;
        network_step_block(gv, &results[1], gv->blocks1, FALSE);

        record_subject_means1(gv, TR_CONTINGENT, results);

        /* Now initialise network then give non-contingent training */
        /* Test for 120 trials where imperative is 1 and 120 when it is 2 */
        /* Imitative stimulus is compatible on 60, incomptible on 60 */

        network_weights_initialise(gv);
        network_set_training_data(gv, TR_NONCONTINGENT);
        network_weights_train(gv, FALSE);

        /* Block 1: Always open: */
        gv->sensory_clamp_settings[0]    = 1;
        gv->imperative_clamp_settings[0] = 1;
        gv->sensory_clamp_settings[1]    = 2;
        gv->imperative_clamp_settings[1] = 1;
        network_step_block(gv, &results[0], gv->blocks1, FALSE);

        /* Block 2: Always close: */
        gv->sensory_clamp_settings[0]    = 2;
        gv->imperative_clamp_settings[0] = 2;
        gv->sensory_clamp_settings[1]    = 1;
        gv->imperative_clamp_settings[1] = 2;
        network_step_block(gv, &results[1], gv->blocks1, FALSE);

        record_subject_means1(gv, TR_NONCONTINGENT, results);

        gv->subject_count1++;

        update_graphics(gv);
    }
}

/******************************************************************************/

double exp2_get_condition_rt_mean(GlobalVariables *gv, TrainingType tt, int condition)
{
    double sum;
    int i;

    sum = 0.0;
    for (i = 0; i < gv->subject_count2; i++) {
        sum += subject_means[(int) tt][condition][i];
    }
    return((gv->subject_count2 > 0) ? (sum / (double) gv->subject_count2) : 0.0);
}

double exp2_get_condition_rt_se(GlobalVariables *gv, TrainingType tt, int condition)
{
    double sum, ssq;
    int i;

    sum = 0.0; ssq = 0.0;
    for (i = 0; i < gv->subject_count2; i++) {
        sum += subject_means[(int) tt][condition][i];
        ssq += subject_means[(int) tt][condition][i] * subject_means[(int) tt][condition][i];
    }
    return((gv->subject_count2 > 1) ? sqrt(((ssq * gv->subject_count2) - (sum * sum)) / (double) (gv->subject_count2 * (gv->subject_count2 - 1))) / (sqrt(gv->subject_count2)) : 0.0);
}

double exp2_get_condition_difference_mean(GlobalVariables *gv, TrainingType tt)
{
    double sum;
    int i;

    sum = 0.0;
    for (i = 0; i < gv->subject_count2; i++) {
        sum += (subject_means[(int) tt][1][i] - subject_means[(int) tt][0][i]);
    }
    return((gv->subject_count2 > 0) ? (sum / (double) gv->subject_count2) : 0.0);
}

double exp2_get_condition_difference_se(GlobalVariables *gv, TrainingType tt)
{
    double sum, ssq, diff;
    int i;

    sum = 0.0; ssq = 0.0;
    for (i = 0; i < gv->subject_count2; i++) {
        diff = (subject_means[(int) tt][1][i] - subject_means[(int) tt][0][i]);
        sum += diff;
        ssq += diff * diff;
    }
    return((gv->subject_count2 > 1) ? sqrt(((ssq * gv->subject_count2) - (sum * sum)) / (double) (gv->subject_count2 * (gv->subject_count2 - 1))) / (sqrt(gv->subject_count2)) : 0.0);
}

/******************************************************************************/

static void record_subject_means2(GlobalVariables *gv, TrainingType tt, SubjectResults results[2])
{
    double sum;
    int c, i;

    for (c = 0; c < NUM_CONDITIONS; c++) {
        if (gv->blocks2 == 0) {
            subject_means[(int) tt][c][gv->subject_count2] = 0;
        }
        else {
            sum = 0.0;
            for (i = 0; i < gv->blocks2; i++) {
                // Assumes trial isn't an error - should check results->response[i][j], 
                sum += results[0].rt[i][c];
                sum += results[1].rt[i][c];
            }
            subject_means[(int) tt][c][gv->subject_count2] = (sum / (2.0 * gv->blocks2));
        }
    }
}

double exp2_fit_rms(GlobalVariables *gv)
{
    double err = 0.0;
    double x;
    int n = 0;

    x = 20.8 - exp2_get_condition_difference_mean(gv, TR_CONTINGENT);
    err += x * x; n++;
    x = 33.6 - exp2_get_condition_difference_mean(gv, TR_NONCONTINGENT);
    err += x * x; n++;
    x = 18.6 - exp2_get_condition_difference_mean(gv, TR_SIGNALLED);
    err += x * x; n++;

    return(sqrt(err / (double) n));
}

double exp2_fit_max(GlobalVariables *gv)
{
    double err = 0.0;
    double x;

    x = 20.8 - exp2_get_condition_difference_mean(gv, TR_CONTINGENT);
    err = MAX(err, fabs(x));
    x = 33.6 - exp2_get_condition_difference_mean(gv, TR_NONCONTINGENT);
    err = MAX(err, fabs(x));
    x = 18.6 - exp2_get_condition_difference_mean(gv, TR_SIGNALLED);
    err = MAX(err, fabs(x));

    return(err);
}

double exp2_fit_psp_score(GlobalVariables *gv, int n)
{
    double means[3];
    double ses[3];
    double df, var_pooled, t, p;
    int score = 0;

    if (gv->subject_count2 > 1) {

        means[0] = exp2_get_condition_difference_mean(gv, TR_CONTINGENT);
        ses[0] = exp2_get_condition_difference_se(gv, TR_CONTINGENT);
        means[1] = exp2_get_condition_difference_mean(gv, TR_NONCONTINGENT);
        ses[1] = exp2_get_condition_difference_se(gv, TR_NONCONTINGENT);
        means[2] = exp2_get_condition_difference_mean(gv, TR_SIGNALLED);
        ses[2] = exp2_get_condition_difference_se(gv, TR_SIGNALLED);

        df = 2 * (gv->subject_count2 - 1);

        if (n == 0) {

            /* Non-Contingent versus contingent: */
            var_pooled = (ses[1] * ses[1] + ses[0] * ses[0]) * gv->subject_count2 * 0.5;
            t = (means[1] - means[0]) / sqrt(var_pooled * 2.0 / gv->subject_count2);
            p = prob_of_t(t, df);
    
            if (p < 0.05) { score = -1; }
            if (p > 0.95) { score = +1; }
	}
        else if (n == 1) {

            /* Non-Contingent versus signalled: */
            var_pooled = (ses[1] * ses[1] + ses[2] * ses[2]) * gv->subject_count2 * 0.5;
            t = (means[1] - means[2]) / sqrt(var_pooled * 2.0 / gv->subject_count2);
            p = prob_of_t(t, df);

            if (p < 0.05) { score = -1; }
            if (p > 0.95) { score = +1; }
	}
        else if (n == 2) {

            /* Signalled versus contingent: */
            var_pooled = (ses[2] * ses[2] + ses[0] * ses[0]) * gv->subject_count2 * 0.5;
            t = (means[2] - means[0]) / sqrt(var_pooled * 2.0 / gv->subject_count2);
            p = prob_of_t(t, df);

            if (p < 0.05) { score = -1; }
            if (p > 0.95) { score = +1; }
	}
    }

    return((double) score);
}

/******************************************************************************/

void exp2_print_analysis(FILE *fp, GlobalVariables *gv)
{
    fprintf(fp, "RMS Error: %5.2f\n", exp2_fit_rms(gv));
    fprintf(fp, "Max Error: %5.2f\n", exp2_fit_max(gv));
}

void exp2_do_task(GlobalVariables *gv)
{
    SubjectResults results[2];
    int s;

    gv->subject_count2 = 0;

    for (s = 0; s < gv->subjects2; s++) {

        /* Initialise network then give contingent training */
        /* Test for 120 trials where imperative is 0 and 120 when it is 1 */
        /* Imitative stimulus is compatible on 60, incomptible on 60 */

        network_weights_initialise(gv);
        network_set_training_data(gv, TR_CONTINGENT);
        network_weights_train(gv, FALSE);

        /* Block 1: Always open: */
        gv->sensory_clamp_settings[0]    = 1;
        gv->imperative_clamp_settings[0] = 1;
        gv->sensory_clamp_settings[1]    = 2;
        gv->imperative_clamp_settings[1] = 1;
        network_step_block(gv, &results[0], gv->blocks2, FALSE);

        /* Block 2: Always close: */
        gv->sensory_clamp_settings[0]    = 2;
        gv->imperative_clamp_settings[0] = 2;
        gv->sensory_clamp_settings[1]    = 1;
        gv->imperative_clamp_settings[1] = 2;
        network_step_block(gv, &results[1], gv->blocks2, FALSE);

        record_subject_means2(gv, TR_CONTINGENT, results);

        /* Now initialise network then give non-contingent training */
        /* Test for 120 trials where imperative is 0 and 120 when it is 1 */
        /* Imitative stimulus is compatible on 60, incomptible on 60 */

        network_weights_initialise(gv);
        network_set_training_data(gv, TR_NONCONTINGENT);
        network_weights_train(gv, FALSE);

        /* Block 1: Always open: */
        gv->sensory_clamp_settings[0]    = 1;
        gv->imperative_clamp_settings[0] = 1;
        gv->sensory_clamp_settings[1]    = 2;
        gv->imperative_clamp_settings[1] = 1;
        network_step_block(gv, &results[0], gv->blocks2, FALSE);

        /* Block 2: Always close: */
        gv->sensory_clamp_settings[0]    = 2;
        gv->imperative_clamp_settings[0] = 2;
        gv->sensory_clamp_settings[1]    = 1;
        gv->imperative_clamp_settings[1] = 2;
        network_step_block(gv, &results[1], gv->blocks2, FALSE);

        record_subject_means2(gv, TR_NONCONTINGENT, results);

        /* Now initialise network then give non-contingent training */
        /* Test for 120 trials where imperative is 0 and 120 when it is 1 */
        /* Imitative stimulus is compatible on 60, incomptible on 60 */

        network_weights_initialise(gv);
        network_set_training_data(gv, TR_SIGNALLED);
        network_weights_train(gv, FALSE);

        /* Block 1: Always open: */
        gv->sensory_clamp_settings[0]    = 1;
        gv->imperative_clamp_settings[0] = 1;
        gv->sensory_clamp_settings[1]    = 2;
        gv->imperative_clamp_settings[1] = 1;
        network_step_block(gv, &results[0], gv->blocks2, FALSE);

        /* Block 2: Always close: */
        gv->sensory_clamp_settings[0]    = 2;
        gv->imperative_clamp_settings[0] = 2;
        gv->sensory_clamp_settings[1]    = 1;
        gv->imperative_clamp_settings[1] = 2;
        network_step_block(gv, &results[1], gv->blocks2, FALSE);

        record_subject_means2(gv, TR_SIGNALLED, results);

        gv->subject_count2++;

        update_graphics(gv);
    }
}

/******************************************************************************/
