/*******************************************************************************

    FILE: 
    AUTHOR: R. Cooper
    CREATION DATE: 
    CONTENTS: 

*******************************************************************************/

#include "cs_maths.h"
#include "lib_maths.h"
#include "stdio.h"

/******************************************************************************/

static double squash(double x)
{
   double y = exp(x);
   return((y - 1.0) / (y + 1.0));
}

static double bounding_function(MathsParameters *mp, double x)
{
    double r = (mp->rest_act - ((mp->max_act+mp->min_act) / 2.0)) * (2.0 / (mp->max_act - mp->min_act));
    double fx, a;

    a = log ((1.0 + r) / (1.0 - r));

    switch (mp->maths_system)  {
	case MT_SIGMOID:
	    fx = x + a;
	    break;

	case MT_TANH:
	    fx = (2.0 * x) + a;
	    break;

	case MT_HALFX:
	    fx = (x / 2.0) + a;
	    break;

	default:
	    fprintf(stderr, "Invalid MathsType in bounding_function\n");
	    return 0.0;
	    break;
    }

    return((squash(fx) * (mp->max_act - mp->min_act) / 2.0) + ((mp->max_act + mp->min_act) / 2.0));
}

/*............................................................................*/

static double inverse_bounding_function(MathsParameters *mp, double x)
{
    double r = (mp->rest_act - ((mp->max_act + mp->min_act) / 2.0)) * (2.0 / (mp->max_act - mp->min_act));
    double t = (x -       ((mp->max_act + mp->min_act) / 2.0)) * (2.0 / (mp->max_act - mp->min_act));
    double a = log ((1.0 + r) / (1.0 - r));
    double b = log ((1.0 + t) / (1.0 - t));
    double c = b - a;
    double fx = 0.0;

    switch (mp->maths_system) {
	case MT_SIGMOID:
	    fx = c;
	    break;

	case MT_TANH:
	    fx = c / 2.0;
	    break;

	case MT_HALFX:
	    fx = c * 2.0;
	    break;

	default:
	    fprintf(stderr, "Invalid MathsType in inverse_bounding_function\n");
	    break;
    }

    return(fx);
}

/*............................................................................*/

static double epsilon_function(MathsParameters *mp, double previous, double input, double x)
{
    double epsilon = 0.0;

    if ((input > 0.0) && (previous < mp->max_act)) {
	epsilon = x * (mp->max_act - previous);
    }
    else if ((input < 0.0) && (previous > mp->min_act)) {
	epsilon = x * (previous - mp->min_act);
    }

    return((mp->persistence * previous) + ((1 - mp->persistence) * mp->rest_act) + epsilon);
}

/*............................................................................*/

double apply_equation(MathsParameters *mp, double previous, double input)
{
    double x;
    double fx = 0.0;

    input += random_normal(0.0, mp->noise);

    switch (mp->maths_system) {

    case MT_SIGMOID: case MT_TANH: case MT_HALFX:
	x = inverse_bounding_function(mp, previous)*mp->persistence+input;
	fx = bounding_function(mp, x);
	break;
    case MT_M_R:
	x = input;
	fx = epsilon_function(mp, previous, input, x);
	break;

    case MT_HOUGHTON:
	x = squash(input);
	fx = epsilon_function(mp, previous, input, x);
	break;

    case MT_NORMALISED:
	fx = (previous * mp->persistence) + input;
	break;

    case MT_COHEN:
        fx = previous * mp->persistence + sigmoid(input) * (1-mp->persistence);
        break;

    case MT_BOUCHER: // NOTE: persistence = (1.0 - leakage);

        // Scale input to make the range sensible; Perhaps not right given LI?
        fx = previous * mp->persistence + input*0.005;
        break;

    default:
	fprintf(stderr, "Invalid MathsType (%d) in apply_equation\n", (int) mp->maths_system);
	break;
    }

    if (fx < mp->min_act) {
        return(mp->min_act);
    }
    else if (fx > mp->max_act) {
        return(mp->max_act);
    }
    else {
        return(fx);
    }
}

/******************************************************************************/
/******************************************************************************/

double initial_activation(MathsParameters *mp, double bias)
{
    if (mp->maths_system == MT_COHEN) { // FIXME
        double rest = sigmoid(bias);
        double input = random_normal(bias, mp->noise);

        return(apply_equation(mp, rest, input));
    }
    else if (mp->maths_system == MT_BOUCHER) {
        return(0.0);
    }
    else {
        double input = random_normal(0.0, mp->noise);

        return(apply_equation(mp, mp->rest_act, input));
    }
}

/******************************************************************************/
/******************************************************************************/
