/******************************************************************************/

#include "lib_maths.h"
#include <stdlib.h>
#include <time.h>

/******************************************************************************/

void random_randomise_seed()
{
    long t0 = (long) time(NULL);
    srand((int) t0);
}

double random_normal(double mean, double sd)
{
    /* This generates a random integer with given mean and standard deviation */

    double r1 = rand() / (double) RAND_MAX;
    double r2 = rand() / (double) RAND_MAX;

    return(mean + sd * sqrt(-2 * log(r1)) * cos(2.0 * M_PI * r2));
}

double random_uniform(double low, double high)
{
    double r = rand() / (double) RAND_MAX;

    return(low + (r * (high - low)));
}

int random_int(int n)
{
    /* This generates a random integer in the range [0, n) */

    return(n * (rand() / (double) RAND_MAX));
}

double squared(double input)
{
    return(input * input);
}

static double a_log_b(double a, double b)
{
    return(a * log(b));
}

double sigmoid(double input)
{
    return(1.0 / (1.0 + exp(-1.0 * input)));
}

double sigmoid_inverse(double input)
{
    if (input >= 1.0) {
        return(sigmoid_inverse(0.9999));
    }
    else if (input <= 0.0) {
        return(sigmoid_inverse(0.0001));
    }
    else {
        return(-log((1.0 / input) - 1.0));
    }
}

double sigmoid_derivative(double input)
{
    double y = sigmoid(input);
    return(y * (1.0 - y));
}

double vector_sum_square_difference(int w, double *v1, double *v2)
{
    double r = 0.0;
    while (w-- > 0) {
        r += squared(v1[w] - v2[w]);
    }
    return(r);
}

double vector_cross_entropy(int w, double *desired, double *actual)
{
    double r = 0.0;
    while (w-- > 0) {
        r -= a_log_b(desired[w], actual[w]) + a_log_b(1-desired[w], 1-actual[w]);
    }
    return(r);
}

double vector_set_variability(int n, int l, double *v)
{
    /* Variability in a set of n vectors, each with l components,   */
    /* located from address v.                                      */

    double variability = 0.0;
    int i, j, count = 0;

    if (n > 1) {
        for (i = 0; i < n-1; i++) {
            for (j = i+1; j < n; j++) {
                variability += sqrt(vector_sum_square_difference(l, &v[i*l], &v[j*l]));
                count++;
            }
        }
    }
    return(variability / (double) count);
}

/*----------------------------------------------------------------------------*/

static int vector_get_index_of_max(double *vector, int l)
{
    int i = --l;

    while (l-- > 0) {
        if (vector[l] > vector[i]) {
            i = l;
        }
    }
    return(i);        
}

void vector_sort(double *vector, int l)
{
    // TO DO: Make this sorting routine more efficient
    while (l-- > 0) {
        int j = vector_get_index_of_max(vector, l+1);
        if (j != l) {
            double tmp = vector[j];
            vector[j] = vector[l];
            vector[l] = tmp;
        }
    }
}

/*----------------------------------------------------------------------------*/

void vector_print(FILE *fp, double *vector, int l)
{
    int j;

    for (j = 0; j < l; j++) {
        fprintf(fp, "%5.3f ", vector[j]);
    }
    fprintf(fp, "\n");
}

/******************************************************************************/

static double power(double x, double y)
{
    return(exp(y * log(x)));
}

static double my_t(int df, double x)
{
    return(1.0 / power(1.0 + (x * x / (double) df), (df + 1.0) / 2.0));
}

double factorial(double n)
{
    if ((n == 0.0) || (n == 1.0)) {
        return(1.0);
    }
    else if (n == -0.5) {
        return(sqrt(PI));
    }
    else if (n == 0.5) {
        return(0.5 * sqrt(PI));
    }
    else if (n > 0) {
        return(n * factorial(n-1));
    }
    else {
        return(0.0);
    }
}

double students_t_constant(int df)
{
    double num = factorial((df-1)/2.0);
    double denom = factorial((df-2)/2.0);

    return(num / (denom * sqrt(df * PI)));
}

double integrate_t_above(double x, int df)
{
    double width, area = 0.0, d;
    double x0, x1, g0, g1;

    x0 = x;
    g0 = my_t(df, x0);
    width = 0.001;
    do {
        x1 = x0 + width;
        g1 = my_t(df, x1);
        area = area + width * (g0 + g1) / 2.0;
        d = g0 - g1;
        x0 = x1;
        g0 = g1;
    } while (d > 1.0e-5);
    width = 0.01;
    do {
        x1 = x0 + width;
        g1 = my_t(df, x1);
        area = area + width * (g0 + g1) / 2.0;
        d = g0 - g1;
        x0 = x1;
        g0 = g1;
    } while (d > 1.0e-7);
    width = 0.1;
    do {
        x1 = x0 + width;
        g1 = my_t(df, x1);
        area = area + width * (g0 + g1) / 2.0;
        d = g0 - g1;
        x0 = x1;
        g0 = g1;
    } while (d > 1.0e-9);
    width = 1.0;
    do {
        x1 = x0 + width;
        g1 = my_t(df, x1);
        area = area + width * (g0 + g1) / 2.0;
        d = g0 - g1;
        x0 = x1;
        g0 = g1;
    } while (d > 1.0e-11);
    width = 10.0;
    do {
        x1 = x0 + width;
        g1 = my_t(df, x1);
        area = area + width * (g0 + g1) / 2.0;
        d = g0 - g1;
        x0 = x1;
        g0 = g1;
    } while (d > 1.0e-13);
    width = 100.0;
    do {
        x1 = x0 + width;
        g1 = my_t(df, x1);
        area = area + width * (g0 + g1) / 2.0;
        d = g0 - g1;
        x0 = x1;
        g0 = g1;
    } while (d > 1.0e-15);
    return(area);
}

double prob_of_t(double t, int df)
{
    if (t < 0.0) {
        return(1.0 - prob_of_t(-t, df));
    }
    else {
        return(students_t_constant(df) * integrate_t_above(t, df));
    }
}

/******************************************************************************/
