#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cs_maths.h"

#define LAYER_WIDTH 3

#undef HEBB
#define RW

typedef struct pattern_list {
    double                input[LAYER_WIDTH];
    double                output[LAYER_WIDTH];
    struct pattern_list  *next;
} PatternList;

static double null_vector[LAYER_WIDTH] = {0.0, 0.0, 0.0};

static double RestAct = 0.1;
static double Treshold = 0.9;

static double sense_node[LAYER_WIDTH], motor_node[LAYER_WIDTH];
static double weights[LAYER_WIDTH][LAYER_WIDTH];

#define LAT_INH  1.0
#define SELF_ACT 0.7

static void free_dataset(PatternList *p)
{
    PatternList *q;

    while (p != NULL) {
        q = p->next;
        free(p);
        p = q;
    }
}

static PatternList *pattern_list_append(PatternList *dataset, double i0, double i1, double i2, double o0, double o1, double o2)
{
    PatternList *node = (PatternList *)malloc(sizeof(PatternList));

    if (node != NULL) {
        node->input[0] = i0;
        node->input[1] = i1;
        node->input[2] = i2;
        node->output[0] = o0;
        node->output[1] = o1;
        node->output[2] = o2;
        node->next = dataset;
        return(node);
    }
    else {
        return(dataset);
    }
}


static PatternList *create_dataset1()
{
    PatternList *dataset = NULL;

// See nothing, do nothing:
    dataset = pattern_list_append(dataset, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0);
// See nothing, do something:
    dataset = pattern_list_append(dataset, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0);
    dataset = pattern_list_append(dataset, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
    dataset = pattern_list_append(dataset, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0);
// See something but don't do it (watching someone else):
    dataset = pattern_list_append(dataset, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0);
    dataset = pattern_list_append(dataset, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0);
    dataset = pattern_list_append(dataset, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0);
// See something and do it:
    dataset = pattern_list_append(dataset, 1.0, 0.0, 0.0, 1.0, 0.0, 0.0);
    dataset = pattern_list_append(dataset, 0.0, 1.0, 0.0, 0.0, 1.0, 0.0);
    dataset = pattern_list_append(dataset, 0.0, 0.0, 1.0, 0.0, 0.0, 1.0);

    return(dataset);
};

static PatternList *create_dataset2()
{
    PatternList *dataset = NULL;

// See nothing, do nothing:
    dataset = pattern_list_append(dataset, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0);
// See something and do something else:
    dataset = pattern_list_append(dataset, 1.0, 0.0, 0.0, 0.0, 1.0, 0.0);
    dataset = pattern_list_append(dataset, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0);
    dataset = pattern_list_append(dataset, 0.0, 0.0, 1.0, 1.0, 0.0, 0.0);

    return(dataset);
}

/******************************************************************************/

static void initialise_network()
{
    int i, j;

    for (i = 0; i < LAYER_WIDTH; i++) {
        for (j = 0; j < LAYER_WIDTH; j++) {
            weights[i][j] = random_normal(0.0, 0.2);
        }
    }
}

/******************************************************************************/

static double calculate_output_energy()
{
    int i, j;
    double e = 0;
    for (i = 0; i < LAYER_WIDTH; i++) {
        for (j = 0; j < LAYER_WIDTH; j++) {
            if (i != j) {
                e += motor_node[i] * motor_node[j] * LAT_INH;
            }
        }
    }
    return(e);
}

static int unique_network_response()
{
    if (0) {
        return(calculate_output_energy() < 0.05);
    }
    else {
        int j = 0, c1 = 0, c2 = 0;
        for (j = 0; j < LAYER_WIDTH; j++) {
            if (motor_node[j] < RestAct) {
                c1++;
            }
            if (motor_node[j] > Treshold) {
                c2++;
            }
	}       
        return((c1 == (LAYER_WIDTH-1)) && (c2 == 1));
    }
}

static void set_network_input(double pattern[LAYER_WIDTH])
{
    int j;

    for (j = 0; j < LAYER_WIDTH; j++) {
        sense_node[j] = pattern[j];
    }
}

static void initialise_network_output()
{
    int j;

    for (j = 0; j < LAYER_WIDTH; j++) {
        motor_node[j] = initial_activation(0.5);
    }
}


static void print_network_state(FILE *fp, int settling_time)
{
    int j;

    for (j = 0; j < LAYER_WIDTH; j++) {
        fprintf(fp, "%4.2f ",  sense_node[j]);
    }
    fprintf(fp, "| ");
    for (j = 0; j < LAYER_WIDTH; j++) {
        fprintf(fp, "%4.2f ",  motor_node[j]);
    }
    if (unique_network_response()) {
        fprintf(fp, "[Settled in %d cycles]\n", settling_time);
    }
    else {
        fprintf(fp, "[Failed to settle after %d cycles]\n", settling_time);
    }
}

static void recirculate_activation()
{
    int i, j;

    for (j = 0; j < LAYER_WIDTH; j++) {
        double input = 0.0;
        /* Feedforward excitation: */
        for (i = 0; i < LAYER_WIDTH; i++) {
            input += sense_node[i] * weights[j][i];
        }
        /* Self excitation (optional): */
        input += (motor_node[j] - RestAct) * SELF_ACT;
        /* Lateral inhibition (optional): */
        for (i = 0; i < LAYER_WIDTH; i++) {
            if (i != j) {
                input -= (motor_node[i] - RestAct) * LAT_INH;
            }
        }
        /* Now recalculate: */
        motor_node[j] = calculate_activation(motor_node[j], input);
    }
}

static int run_network_to_quiescence(int max_cycles)
{
    int i = 0;

    while ((i < max_cycles) && !unique_network_response()) {
        recirculate_activation(); i++;
    }
    return(i);
}

static void test_network(FILE *fp, char *message)
{
    int i, j, c;

    fprintf(fp, "Testing network (%s)\n", message);

    fprintf(fp, "Weight matrix:\n");
    for (i = 0; i < LAYER_WIDTH; i++) {
        fprintf(fp, "    ");
        for (j = 0; j < LAYER_WIDTH; j++) {
            fprintf(fp, "%8.3f ", weights[j][i]);
        }
        fprintf(fp, "\n");
    }
    fprintf(fp, "\n");


    set_network_input(null_vector);
    initialise_network_output();
    c = run_network_to_quiescence(1000);
    print_network_state(fp, c);

    for (i = 0; i < LAYER_WIDTH; i++) {
        set_network_input(null_vector);
        sense_node[i] = 1.0;
        initialise_network_output();
        c = run_network_to_quiescence(1000);
        print_network_state(fp, c);
    }

    fprintf(fp, "\n");
}

#ifdef HEBB
static void train_network(PatternList *training_data)
{
    double delta_weights[LAYER_WIDTH][LAYER_WIDTH];
    double lr = 0.05;
    PatternList *k;
    int i, j;

    for (i = 0; i < LAYER_WIDTH; i++) {
        for (j = 0; j < LAYER_WIDTH; j++) {
            delta_weights[j][i] = 0.0;
        }
    }

    for (k = training_data; k != NULL; k = k->next) {
        /* Set the network to the pattern: */
        for (i = 0; i < LAYER_WIDTH; i++) {
            sense_node[i] = k->input[i];
            motor_node[i] = k->output[i];
        }
        /* Now calculate delta w: */
        for (i = 0; i < LAYER_WIDTH; i++) {
            for (j = 0; j < LAYER_WIDTH; j++) {
                delta_weights[j][i] += sense_node[i] * motor_node[j];
            }
        }
    }

    for (i = 0; i < LAYER_WIDTH; i++) {
        for (j = 0; j < LAYER_WIDTH; j++) {
            weights[j][i] += lr * delta_weights[j][i];
        }
    }

}
#endif

#ifdef RW
static void train_network(PatternList *training_data)
{
    double delta_weights[LAYER_WIDTH][LAYER_WIDTH];
    double lr = 0.01;
    PatternList *k;
    int i, j;

    for (i = 0; i < LAYER_WIDTH; i++) {
        for (j = 0; j < LAYER_WIDTH; j++) {
            delta_weights[j][i] = 0.0;
        }
    }

    for (k = training_data; k != NULL; k = k->next) {
        /* Set the network input to the pattern: */
        for (i = 0; i < LAYER_WIDTH; i++) {
            sense_node[i] = k->input[i];
        }
        run_network_to_quiescence(1000);

        /* Now calculate delta w: */
        for (i = 0; i < LAYER_WIDTH; i++) {
            for (j = 0; j < LAYER_WIDTH; j++) {
                delta_weights[j][i] += sense_node[i] * (k->output[j]-motor_node[j]);
            }
        }
    }

    for (i = 0; i < LAYER_WIDTH; i++) {
        for (j = 0; j < LAYER_WIDTH; j++) {
            weights[j][i] += lr * delta_weights[j][i];
        }
    }

}
#endif

int main(int argc, char **argv)
{
    PatternList *dataset;
    int i;

    randomise();
    initialise_network();

    dataset = create_dataset1();
    test_network(stdout, "before training");
    for (i = 0; i < 1000; i++) {
        train_network(dataset);
    }
    free_dataset(dataset);
    test_network(stdout, "after 1000 epochs standard training");

    dataset = create_dataset2();
    for (i = 0; i < 250; i++) {
        train_network(dataset);
    }
    free_dataset(dataset);
    test_network(stdout, "after 250 epochs crossed training");

    exit(0);
}
