
#include "stdio.h"
#include "stdlib.h"
#include "lib_maths.h"
#include "cs_maths.h"

#include "src_model.h"

/* Used for RW learning (version 2) only: */
#define RW_BASELINE 0.5

// Standard deviation of IM strength across subjects:
//#define SUBJECT_NOISE 1.0
#define SUBJECT_NOISE 0.00001
// Standard deviation of learning rate noise across subjects:
# define LR_NOISE 0.00000001

typedef enum ctx_type {CTX_C1_OPEN, CTX_C1_CLOSE, CTX_C2_OPEN, CTX_C2_CLOSE} CtxType;

Assoc association[NUM_MOTOR][NUM_SENSE+NUM_CONTEXT];

/******************************************************************************/

char *lr_name[NUM_LR] = {
    "Hebb",
    "Quasi-Hebb",
    "R-W",
    "R-W (with baseline)"
};

/******************************************************************************/

static void free_training_data(PatternList *p)
{
    PatternList *q;

    while (p != NULL) {
        q = p->next;
        free(p);
        p = q;
    }
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

static int pattern_list_length(PatternList *list)
{
    int n = 0;
    while (list != NULL) {
        list = list->next;
        n++;
    }
    return(n);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

static PatternList *pattern_list_remove(PatternList *list, int k, PatternList **tmp)
{
    if (list == NULL) {
        *tmp = NULL;
        return(NULL);
    }
    else if (k == 0) {
        *tmp = list;
        return(list->next);
    }
    else {
        PatternList *prev = list;
        while (--k > 0) {
            prev = prev->next;
        }
        *tmp = prev->next;
        prev->next = (*tmp)->next;
        return(list);
    }
}

#ifdef DEBUG
    static void pattern_list_print(FILE *fp, char *label, PatternList *result)
{
    fprintf(fp, "%s: ", label);
    while (result != NULL) {
        fprintf(fp, "%p ", result);
        result = result->next;
    }
    fprintf(fp, "\n");
}
#endif

static PatternList *pattern_list_shuffle(PatternList *list)
{
    PatternList *result = NULL, *tmp;
    int j, k, n;

    n = pattern_list_length(list);

    for (j = n; j > 0; j--) {
        k = random_int(j);
        // Now remove k-th element from list and prepend it to result;
        list = pattern_list_remove(list, k, &tmp);
        tmp->next = result;
        result = tmp;
    }

    return(result);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

static PatternList *pattern_list_append(GlobalVariables *gv, PatternList *dataset, double i0, double i1, CtxType c, double p0, double p1, double o0, double o1)
{
    PatternList *node = (PatternList *)malloc(sizeof(PatternList));

    // CORRECT FOR NORMAL TRIALS, INCORECT FOR SIGNALLED TRIALS

    if (node != NULL) {
        node->input[0] = i0;
        node->input[1] = i1;
        node->imperative[0] = p0;
        node->imperative[1] = p1;
        if (c == CTX_C1_OPEN) {
            node->context[0] = 1.0;
            node->context[1] = 1.0;
            node->context[2] = 0.0;
            node->context[3] = 0.0;
        }
        else if (c == CTX_C1_CLOSE) {
            node->context[0] = 0.0;
            node->context[1] = 1.0;
            node->context[2] = 0.0;
            node->context[3] = 1.0;
        }
        else if (c == CTX_C2_OPEN) {
            node->context[0] = 1.0;
            node->context[1] = 0.0;
            node->context[2] = 1.0;
            node->context[3] = 0.0;
        }
        else if (c == CTX_C2_CLOSE) {
            node->context[0] = 0.0;
            node->context[1] = 0.0;
            node->context[2] = 1.0;
            node->context[3] = 1.0;
        }
        node->output[0] = o0;
        node->output[1] = o1;
        node->next = dataset;
        return(node);
    }
    else {
        return(dataset);
    }
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

static PatternList *create_dataset(GlobalVariables *gv)
{
    PatternList *dataset = NULL;
    int i;

    // Patterns are: Input action; Imperative stimulus; Motor response
    // Each block consists of 36 of each type of pattern

    for (i = 0; i < 36; i++) {
        switch (gv->training_type) {
            case TR_CONTINGENT: { // Sensory is always reversed
                // Sensory (open), Imperative (close), Output (close);
                dataset = pattern_list_append(gv, dataset, 1.0, 0.0, CTX_C1_CLOSE, 0.0, 1.0, 0.0, 1.0);
                // Sensory (close), Imperative (open), Output (open);
                dataset = pattern_list_append(gv, dataset, 0.0, 1.0, CTX_C1_OPEN, 1.0, 0.0, 1.0, 0.0);
                break;
            }
            case TR_NONCONTINGENT: { // Sensory is reversed or ignored
                // Sensory (open), Imperative (close), Output (close);
                dataset = pattern_list_append(gv, dataset, 1.0, 0.0, CTX_C1_CLOSE, 0.0, 1.0, 0.0, 1.0);
                // Sensory (close), Imperative (open), Output (open);
                dataset = pattern_list_append(gv, dataset, 0.0, 1.0, CTX_C1_OPEN, 1.0, 0.0, 1.0, 0.0);
                // Sensory (none), Imperative (close), Output (close);
                dataset = pattern_list_append(gv, dataset, 0.0, 0.0, CTX_C1_CLOSE, 0.0, 1.0, 0.0, 1.0);
                // Sensory (none), Imperative (open), Output (open);
                dataset = pattern_list_append(gv, dataset, 0.0, 0.0, CTX_C1_OPEN, 1.0, 0.0, 1.0, 0.0);
                break;
            }
            case TR_SIGNALLED: { // Sensory is reversed or ignored with a signal
                // Sensory (open), Imperative (close), Output (close);
                dataset = pattern_list_append(gv, dataset, 1.0, 0.0, CTX_C1_CLOSE, 0.0, 1.0, 0.0, 1.0);
                // Sensory (close), Imperative (open), Output (open);
                dataset = pattern_list_append(gv, dataset, 0.0, 1.0, CTX_C1_OPEN, 1.0, 0.0, 1.0, 0.0);
                // Sensory (none), Imperative (close), Output (close);
                dataset = pattern_list_append(gv, dataset, 0.0, 0.0, CTX_C2_CLOSE, 0.0, 1.0, 0.0, 1.0);
                // Sensory (none), Imperative (open), Output (open);
                dataset = pattern_list_append(gv, dataset, 0.0, 0.0, CTX_C2_OPEN, 1.0, 0.0, 1.0, 0.0);
                break;
            }
            case TR_OPEN_CLOSE: { // Sensory reversed: 1 trial
                // Sensory (open), Imperative (close), Output (close);
                dataset = pattern_list_append(gv, dataset, 1.0, 0.0, CTX_C1_CLOSE, 0.0, 1.0, 0.0, 1.0);
                break;
            }
            case TR_CLOSE_OPEN: { // Sensory reversed: 1 trial
                // Sensory (close), Imperative (open), Output (open);
                dataset = pattern_list_append(gv, dataset, 0.0, 1.0, CTX_C1_OPEN, 1.0, 0.0, 1.0, 0.0);
                break;
            }
            case TR_CLOSE: { // Sensory ignored: 1 trial
                // Sensory (none), Imperative (close), Output (close);
                dataset = pattern_list_append(gv, dataset, 0.0, 0.0, CTX_C2_CLOSE, 0.0, 1.0, 0.0, 1.0);
                break;
            }
            case TR_OPEN: { // Sensory ignored: 1 trial
                // Sensory (none), Imperative (open), Output (open);
                dataset = pattern_list_append(gv, dataset, 0.0, 0.0, CTX_C2_OPEN, 1.0, 0.0, 1.0, 0.0);
                break;
            }
            default: {
            }
        }
    }
    return(dataset);
};

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void network_set_training_data(GlobalVariables *gv, TrainingType tt)
{
    if (gv->training_data == NULL) {
        free_training_data(gv->training_data);
    }

    gv->training_type = tt;
    gv->training_data = create_dataset(gv);

    if (gv->print_output) {
        fprintf(stdout, "Training data reset: %d trials per block\n", pattern_list_length(gv->training_data));
    }
}

/******************************************************************************/

static void network_nodes_initialise(GlobalVariables *gv)
{
    gv->sensory_node[0].colour = 3;
    gv->sensory_node[0].maths = &(gv->s_maths);
    gv->sensory_node[1].colour = 4;
    gv->sensory_node[1].maths = &(gv->s_maths);

    gv->context_node[0].colour = 3;
    gv->context_node[0].maths = &(gv->c_maths);
    gv->context_node[1].colour = 2;
    gv->context_node[1].maths = &(gv->c_maths);
    gv->context_node[2].colour = 2;
    gv->context_node[2].maths = &(gv->c_maths);
    gv->context_node[3].colour = 4;
    gv->context_node[3].maths = &(gv->c_maths);

    gv->imperative_node[0].colour = 3;
    gv->imperative_node[0].maths = &(gv->d_maths);
    gv->imperative_node[1].colour = 4;
    gv->imperative_node[1].maths = &(gv->d_maths);

    gv->motor_node[0].colour = 3;
    gv->motor_node[0].maths = &(gv->r_maths);
    gv->motor_node[1].colour = 4;
    gv->motor_node[1].maths = &(gv->r_maths);
}

/******************************************************************************/

GlobalVariables *globals_create()
{
    GlobalVariables *gv = (GlobalVariables *)malloc(sizeof(GlobalVariables));

    if (gv != NULL) {
        gv->d_maths.maths_system = MT_COHEN; // MT_BOUCHER;
        gv->d_maths.persistence = 0.990; // 1.00;
        gv->s_maths.maths_system = MT_COHEN; // MT_BOUCHER;
        gv->s_maths.persistence = 0.960; // 1.00;
        gv->r_maths.maths_system = MT_COHEN; // MT_BOUCHER;
        gv->r_maths.persistence = 0.990; // 1.00;
        gv->c_maths.maths_system = MT_COHEN; // MT_BOUCHER;
        gv->c_maths.persistence = 0.990; // 1.00;
        gv->sensory_bias = -2.00;
        gv->imperative_bias = -2.00;
        gv->context_bias = -2.00;
        gv->response_bias = -6.00;
        gv->sensory_input_strength = 5.0;
        gv->imperative_input_strength = 5.0;
        gv->context_input_strength[0] =  5.0;  // 7.50;
        gv->context_input_strength[1] =  5.0;  // 7.50;
        gv->context_input_strength[2] = 50.0;  // 75.0;
        gv->context_input_strength[3] =  5.0;  // 7.50;
        gv->sensory_li = 0.0;
        gv->response_li = 0.0;

        gv->selection = ST_THRESHOLD;
        gv->default_response_threshold = 0.80;
        gv->habituation_threshold = 0.90; // Was 0.80 in previous model
        gv->strength_cm =  0.0;
        gv->strength_sm =  4.0;
        gv->strength_im = 10.0;

        gv->d_maths.noise =    2.00;
        gv->d_maths.min_act =  0.00;
        gv->d_maths.max_act =  1.00;
        gv->d_maths.rest_act = 0.00; // initial_activation(&(gv->d_maths), gv->imperative_bias);

        gv->s_maths.noise =    2.00;
        gv->s_maths.min_act =  0.00;
        gv->s_maths.max_act =  1.00;
        gv->s_maths.rest_act = 0.00; // initial_activation(&(gv->s_maths), gv->sensory_bias);

        gv->r_maths.noise =    2.00;
        gv->r_maths.min_act =  0.00;
        gv->r_maths.max_act =  1.00;
        gv->r_maths.rest_act = 0.00; // initial_activation(&(gv->r_maths), gv->response_bias);

        gv->c_maths.noise =    2.00;
        gv->c_maths.min_act =  0.00;
        gv->c_maths.max_act =  1.00;
        gv->c_maths.rest_act = 0.00; // initial_activation(&(gv->c_maths), gv->context_bias);

        gv->sensory_habituation = TRUE;
        gv->context_habituation = TRUE;
        gv->dynamic_response_threshold = FALSE;

        gv->training_type = TR_CONTINGENT;
        gv->training_data = create_dataset(gv);
        gv->learning_rule = LR_RW_A;
        gv->target_value = 10.0;
        gv->weight_decay = 0.000;
        gv->learning_rate = 0.0040;

        network_nodes_initialise(gv);
        network_weights_initialise(gv);

        gv->imitative_delay   =   80;

        // SD of imperative-to-motor connections (motivation) across subjects; 1.0 seems ok.
        gv->subject_im_sd = SUBJECT_NOISE;

        gv->print_output = FALSE;
        gv->cycle = 0;
        gv->i_clamp = 0;
        gv->s_clamp = 0;
        gv->epochs = 6;   // Six training blocks (2 * 36 or 4 * 36 trials) in each experiment
        gv->blocks1 = 60;   // 2 testing blocks, each of 120 trials (60 compatible, 60 incompatible) completed during session 3
        gv->blocks2 = 60;   // As for experiment 1
        gv->subjects1 = 12; // 12 subjects in each condition of CDH exp 1
        gv->subjects2 = 12; // 12 subjects in each condition of CDH exp 2
        gv->subject_count1 = 0;
        gv->subject_count2 = 0;
        gv->soa = 0;

        gv->window = NULL;
        gv->canvas_network = NULL;
        gv->canvas_act_profiles = NULL;
        gv->canvas_weight_profiles = NULL;
        gv->canvas1 = NULL;
        gv->c_and_h_scale = 1.0;
    }
    return(gv);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

void globals_destroy(GlobalVariables *gv)
{
    if (gv != NULL) {
        if (gv->training_data != NULL) {
            free_training_data(gv->training_data);
            gv->training_data = NULL;
        }
        free(gv);
    }
}

/******************************************************************************/

static void network_set_clamps_off(GlobalVariables *gv)
{
    int i;

    for (i = 0; i < NUM_IMPERATIVE; i++) {
        gv->imperative_node[i].clamped = FALSE;
    }
    for (i = 0; i < NUM_SENSE; i++) {
        gv->sensory_node[i].clamped = FALSE;
    }
    for (i = 0; i < NUM_CONTEXT; i++) {
        gv->context_node[i].clamped = FALSE;
    }
    for (i = 0; i < NUM_MOTOR; i++) {
        gv->motor_node[i].clamped = FALSE;
    }
}

static void network_set_sensory_imitative_clamps(GlobalVariables *gv)
{
    if (gv->s_clamp == 1) {
        gv->sensory_node[0].clamped = TRUE;
    }
    else if (gv->s_clamp == 2) {
        gv->sensory_node[1].clamped = TRUE;
    }
}

static void network_set_imperative_clamps(GlobalVariables *gv)
{
    /* Set the imperative and context nodes that are no during testing */
    /* Note that context nodes corresponding to the imperative stimulus from */
    /* training are not set as they are not present in testing. */
    if (gv->i_clamp == 1) {
        gv->imperative_node[0].clamped = TRUE;
        gv->imperative_node[1].clamped = FALSE;
    }
    else if (gv->i_clamp == 2) {
        gv->imperative_node[0].clamped = FALSE;
        gv->imperative_node[1].clamped = TRUE;
    }

    /* Context nodes for "1" and "2" are always off during testing */
    gv->context_node[0].clamped = FALSE;
    gv->context_node[3].clamped = FALSE;
    /* Context node "C1" (standard warning stimulus) is always on during testing: */
     gv->context_node[1].clamped = TRUE;
    /* Context node "C2" (signalled warning stimulus) is always off during testing: */
    gv->context_node[2].clamped = FALSE;
}

void update_world(GlobalVariables *gv)
{
    if (gv->cycle == 0) {
        network_set_clamps_off(gv);
    }
    if (gv->cycle == (TIME_ZERO - gv->soa + gv->imitative_delay)) {
        network_set_sensory_imitative_clamps(gv);
    }
    if (gv->cycle == TIME_ZERO) {
        network_set_imperative_clamps(gv);
    }
}

/******************************************************************************/

int sensory_competitor(int i)
{
    switch (i) {
        case 0: { return(1); break; }
        case 1: { return(0); break; }
        default: { return(-1); break; }
    }
}

int context_competitor(int i)
{
    switch (i) {
        case 0: { return(1); break; }
        case 1: { return(0); break; }
        default: { return(-1); break; }
    }
}

static void recirculate_activation(GlobalVariables *gv)
{
    double input;
    int i, j;

    for (i = 0; i < NUM_CONTEXT; i++) {
        input = gv->context_node[i].bias;
        // Sensory lateral inhibition:
        if ((j = context_competitor(i)) >= 0) {
            input -= (gv->context_node[j].activation[gv->cycle] - gv->context_node[j].maths->rest_act) * gv->sensory_li;
	}

        /* Habituation of context units: */
        if (gv->context_habituation) {
            if (gv->context_node[i].clamped) {
                if (gv->context_node[i].activation[gv->cycle] > gv->habituation_threshold) {
                    gv->context_node[i].habituated = TRUE;
                }
                else if (!gv->context_node[i].habituated) {
                    input += gv->context_input_strength[i];
                }
            }
        }
        else {
        /* Add clamped input: */
            if (gv->context_node[i].clamped) {
                input += gv->context_input_strength[i];
            }
        }
        gv->context_node[i].activation[gv->cycle+1] = apply_equation(gv->context_node[i].maths, gv->context_node[i].activation[gv->cycle], input);
    }

    for (i = 0; i < NUM_IMPERATIVE; i++) {
        input = gv->imperative_node[i].bias;
        if (gv->imperative_node[i].clamped) {
            input += gv->imperative_input_strength;
        }

        /* Lateral inhibition (optional): */
        for (j = 0; j < NUM_IMPERATIVE; j++) {
            if (i != j) {
                input -= (gv->imperative_node[j].activation[gv->cycle] - gv->imperative_node[j].maths->rest_act) * gv->sensory_li;
            }
        }

        gv->imperative_node[i].activation[gv->cycle+1] = apply_equation(gv->imperative_node[i].maths, gv->imperative_node[i].activation[gv->cycle], input);
    }

    for (i = 0; i < NUM_SENSE; i++) {
        input = gv->sensory_node[i].bias;
        // Sensory lateral inhibition:
        if ((j = sensory_competitor(i)) >= 0) {
            input -= (gv->sensory_node[j].activation[gv->cycle] - gv->sensory_node[j].maths->rest_act) * gv->sensory_li;
	}

        /* Habituation of sensory units: */
        if (gv->sensory_habituation) {
            if (gv->sensory_node[i].clamped) {
                if (gv->sensory_node[i].activation[gv->cycle] > gv->habituation_threshold) {
                    gv->sensory_node[i].habituated = TRUE;
                }
                else if (!gv->sensory_node[i].habituated) {
                    input += gv->sensory_input_strength;
                }
            }
        }
        else {
        /* Add clamped input: */
            if (gv->sensory_node[i].clamped) {
                input += gv->sensory_input_strength;
            }
        }
        gv->sensory_node[i].activation[gv->cycle+1] = apply_equation(gv->sensory_node[i].maths, gv->sensory_node[i].activation[gv->cycle], input);
    }

    for (j = 0; j < NUM_MOTOR; j++) {
        input = gv->motor_node[j].bias;

        /* Feedforward excitation from imperative stimulus: */
        for (i = 0; i < NUM_IMPERATIVE; i++) {
            /* Only feedforward activation above rest: */
            input += (gv->imperative_node[i].activation[gv->cycle] - gv->imperative_node[i].maths->rest_act) * WEIGHT_IM(j, i);
        }

        /* Feedforward excitation from sensory input: */
        for (i = 0; i < NUM_SENSE; i++) {
            /* Only feedforward activation above rest: */
            input += (gv->sensory_node[i].activation[gv->cycle] - gv->sensory_node[i].maths->rest_act) * WEIGHT_SM(j, i);
        }

        /* Feedforward excitation from context input: */
        for (i = 0; i < NUM_CONTEXT; i++) {
            /* Only feedforward activation above rest: */
            input += (gv->context_node[i].activation[gv->cycle] - gv->context_node[i].maths->rest_act) * WEIGHT_CM(j, i);
        }

        /* Lateral inhibition (optional): */
        for (i = 0; i < NUM_MOTOR; i++) {
            if (i != j) {
                input -= (gv->motor_node[i].activation[gv->cycle] - gv->motor_node[i].maths->rest_act) * gv->response_li;
            }
        }

        /* Now recalculate: */
        gv->motor_node[j].activation[gv->cycle+1] = apply_equation(gv->motor_node[j].maths, gv->motor_node[j].activation[gv->cycle], input);
    }

    /* Now check if a response should occur: */
    switch (gv->selection) {
        case ST_THRESHOLD: {
            int r = 0;
            for (i = 1; i < NUM_MOTOR; i++) {
                if (gv->motor_node[i].activation[gv->cycle] > gv->motor_node[r].activation[gv->cycle]) {
                    r = i;
                }
            }
            if (gv->motor_node[r].activation[gv->cycle] >= gv->actual_response_threshold) {
                gv->motor_node[r].fired = TRUE;
            }
            break;
        }
        case ST_GAP: {
            double gap = 0.0;
            int r = 0;
            for (i = 1; i < NUM_MOTOR; i++) {
                if (gv->motor_node[i].activation[gv->cycle] > gv->motor_node[r].activation[gv->cycle]) {
                    gap = gv->motor_node[i].activation[gv->cycle] - gv->motor_node[r].activation[gv->cycle];
                    r = i;
                }
                else if ((gap == 0) || (gv->motor_node[i].activation[gv->cycle] > (gv->motor_node[r].activation[gv->cycle] - gap))) {
                    gap = gv->motor_node[r].activation[gv->cycle] - gv->motor_node[i].activation[gv->cycle];
                }
            }
            if (gap >= gv->actual_response_threshold) {
                gv->motor_node[r].fired = TRUE;
            }
            break;
        }
        case ST_PROBABILISTIC: {
            for (i = 0; i < NUM_MOTOR; i++) {
                if (gv->motor_node[i].activation[gv->cycle+1] / 25.0 > random_uniform(0.0, 1.0)) {
                    gv->motor_node[i].fired = TRUE;
                }
            }
            break;
        }
        default: {
            break;
        }
    }

    gv->cycle = gv->cycle + 1;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

int network_response(GlobalVariables *gv)
{
    int i;

    for (i = 0; i < NUM_MOTOR; i++) {
        if (gv->motor_node[i].fired) {
            return(i);
        }
    }
    return(-1);
}

/*----------------------------------------------------------------------------*/

//static int network_run_to_quiescence(GlobalVariables *gv)
//{
//    while ((network_response(gv) < 0) && (gv->cycle < (MAX_CYCLES-1))) {
//        recirculate_activation(gv);
//    }
//    return(network_response(gv));
//}

/******************************************************************************/

void network_weights_set_im(GlobalVariables *gv)
{
    int i, j;
    double motivation;

    motivation = gv->strength_im + random_uniform(0.0, gv->subject_im_sd);

    for (i = 0; i < NUM_IMPERATIVE; i++) {
        for (j = 0; j < NUM_MOTOR; j++) {
            if (i == j) {
                WEIGHT_IM(j, i) = motivation;
            }
            else {
                WEIGHT_IM(j, i) = 0.0;
            }
        }
    }
}

void network_weights_initialise(GlobalVariables *gv)
{
    int i, j, k;

    for (i = 0; i < NUM_SENSE; i++) {
        for (j = 0; j < NUM_MOTOR; j++) {
            if (i == j) {
                WEIGHT_SM(j, i) = gv->strength_sm;
            }
            else {
                WEIGHT_SM(j, i) = 0.0;
            }
        }
    }

    /* First set all context weights to zero: */
    for (i = 0; i < NUM_CONTEXT; i++) {
        for (j = 0; j < NUM_MOTOR; j++) {
            WEIGHT_CM(j, i) = 0.0;
        }
    }

    /* Now set the ones that should be non-zero: */
    WEIGHT_CM(0, 0) = gv->strength_cm;
    WEIGHT_CM(0, 1) = gv->strength_cm; // STANDARD
    WEIGHT_CM(1, 1) = gv->strength_cm; // STANDARD
    WEIGHT_CM(0, 2) = gv->strength_cm; // SIGNALLED
    WEIGHT_CM(1, 2) = gv->strength_cm; // SIGNALLED
    WEIGHT_CM(1, 3) = gv->strength_cm;
    /* Actually, the above are *all* zero because strength_cm is zero*/

    /* Now set association matrix. This is a record of the weight change    */
    /* during learning, used only to graph the result.                      */

    for (i = 0; i < NUM_SENSE; i++) {
        for (j = 0; j < NUM_MOTOR; j++) {
            association[j][i].colour = gv->sensory_node[i].colour;
            for (k = 0; k < gv->epochs; k++) {
                association[j][i].strength[k] = gv->strength_sm;
            }
	}
    }

    for (i = NUM_SENSE; i < (NUM_SENSE+NUM_CONTEXT); i++) {
        for (j = 0; j < NUM_MOTOR; j++) {
            association[j][i].colour = gv->context_node[i-NUM_SENSE].colour;
            for (k = 0; k < gv->epochs; k++) {
                association[j][i].strength[k] = gv->strength_cm;
            }
	}
    }

    network_weights_set_im(gv);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/

static void network_weights_train_one_epoch(GlobalVariables *gv, int n)
{
    double delta_weight[NUM_MOTOR][NUM_SENSE+NUM_CONTEXT];
    PatternList *k;
    int i, j;

    int tmp_id = gv->imitative_delay;
    int tmp_sc = gv->s_clamp;
    int tmp_ic = gv->i_clamp;

    gv->imitative_delay = 0;
    gv->s_clamp = 0;
    gv->i_clamp = 0;

    /* Shuffle the training data: */
    gv->training_data = pattern_list_shuffle(gv->training_data);
    
    for (k = gv->training_data; k != NULL; k = k->next) {

        /* Calculate and then apply delta w for each pattern: */

        if (gv->print_output) {
            fprintf(stdout, "Exemplar: %3.1f %3.1f -> %3.1f %3.1f\n", k->input[0], k->input[1], k->output[0], k->output[1]);
	}
        switch (gv->learning_rule) {
            case LR_HEBB: {
                for (j = 0; j < NUM_MOTOR; j++) {
                    for (i = 0; i < NUM_SENSE; i++) {
                        delta_weight[j][i] = k->input[i] * k->output[j];
                    }
                    for (i = 0; i < NUM_CONTEXT; i++) {
                        delta_weight[j][NUM_SENSE+i] = k->context[i] * k->output[j];
                    }
                }

                /* Scale deltas if we have an asymptote: */
                if (gv->target_value > 0.0) {
                    for (j = 0; j < NUM_MOTOR; j++) {
                        for (i = 0; i < NUM_SENSE+NUM_CONTEXT; i++) {
                            delta_weight[j][i] = delta_weight[j][i] * (gv->target_value-gv->weight[j][i]) / gv->target_value;
                        }
                    }
                }

                break;
            }
            /*---- Hebbian + Anti-Hebbian learning */
            case LR_QEBB: { /* Fudged Hebb: Also learn inhibitory weights */
                /* Calculate delta w for this pattern: */
                for (j = 0; j < NUM_MOTOR; j++) {
                    for (i = 0; i < NUM_SENSE; i++) {
                        if ((k->input[i] > 0.1) && (k->output[j] > 0.1)) {
                            // S present and R is present, so strengthen
                            delta_weight[j][i] = k->input[i] * k->output[j];
                        }
                        else if ((k->input[i] < 0.1) && (k->output[j] > 0.1)) {
                            // S absent and R present, so weaken
                            delta_weight[j][i] = -k->output[j];
                        }
                        else if ((k->input[i] > 0.1) && (k->output[j] < 0.1)) {
                            // S present and R is absent, so weaken
                            delta_weight[j][i] = -k->input[i];
                        }
                        else {
                            // S absent and R absent
                            delta_weight[j][i] = 0.0;
                        }
                    }
                    for (i = 0; i < NUM_CONTEXT; i++) {
                        if ((k->context[i] > 0.1) && (k->output[j] > 0.1)) {
                            // S present and R is present, so strengthen
                            delta_weight[j][NUM_SENSE+i] = k->context[i] * k->output[j];
                        }
                        else if ((k->context[i] < 0.1) && (k->output[j] > 0.1)) {
                            // S absent and R present, so weaken
                            delta_weight[j][NUM_SENSE+i] = -k->output[j];
                        }
                        else if ((k->context[i] > 0.1) && (k->output[j] < 0.1)) {
                            // S present and R is absent, so weaken
                            delta_weight[j][NUM_SENSE+i] = -k->context[i];
                        }
                        else {
                            // S absent and R absent
                            delta_weight[j][NUM_SENSE+i] = 0.0;
                        }
    		    }
                }

                /* Scale deltas if we have an asymptote: */
                if (gv->target_value > 0.0) {
                    for (j = 0; j < NUM_MOTOR; j++) {
                        for (i = 0; i < NUM_SENSE+NUM_CONTEXT; i++) {
                            delta_weight[j][i] = delta_weight[j][i] * (gv->target_value-gv->weight[j][i]) / gv->target_value;
                        }
                    }
                }

                break;
            }
            /*---- Rescorla Wagner learning (implementation 1) */
            case LR_RW_A: {
                for (j = 0; j < NUM_MOTOR; j++) {
                    double v, err, target, net_input;

                    target = k->output[j] * gv->target_value;

                    v = 0.0;
                    net_input = 0.0;
                    for (i = 0; i < NUM_SENSE; i++) {
                        v += WEIGHT_SM(j, i) * k->input[i] * gv->sensory_input_strength;
                        net_input += k->input[i] * gv->sensory_input_strength;
                    }
                    for (i = 0; i < NUM_CONTEXT; i++) {
                        v += WEIGHT_CM(j, i) * k->context[i] * gv->context_input_strength[i];
                        net_input += k->context[i] * gv->context_input_strength[i];
                    }
// "target" will either be zero or target_value. This is the value that we
// want to get when we add all the weighted inputs to the given output node.
// Note that we don't want this just to be one, since it is actually the weighted
// input that is put through a squashing function. Perhaps we want this to be
// equal to the inverse of the squashing function applied to the response
// threshold? No - this isn't a sufficiently big target. It actually needs to
// be quite a large number to drive learning. 

                    err = target - v;

                    if ((target > 0) && (err < 0)) {
                        if (gv->print_output) {
                            fprintf(stdout, "OVERPREDICTION: E%3d; ", n);
                            fprintf(stdout, "Node: %d; Target: %f; Actual: %f\n", j, target, v);
                        }
                    }

                    if (gv->print_output) {
                        fprintf(stdout, "Output unit %d:\n    V: %f; LAMBDA: %f; Error: %f\n", j, v, target, err);
                    }
                    for (i = 0; i < NUM_SENSE; i++) {
                        delta_weight[j][i] = (net_input < 0.00001) ? 0.0 : err * k->input[i] * gv->sensory_input_strength / net_input;
                        if (gv->print_output) {
                            fprintf(stdout, "    Delta sensory weight %d -> %d: %f\n", i, j, delta_weight[j][i]);
                        }
                    }
                    for (i = 0; i < NUM_CONTEXT; i++) {
                        delta_weight[j][NUM_SENSE+i] = (net_input < 0.00001) ? 0.0 : err * k->context[i] * gv->context_input_strength[i] / net_input;
                        if (gv->print_output) {
                            fprintf(stdout, "    Delta context weight %d -> %d: %f\n", i, j, delta_weight[j][NUM_SENSE+i]);
                        }
                    }
                }
                break;
            }
            /*---- Rescorla Wagner learning (implementation 2) */
            case LR_RW_B: {
                for (j = 0; j < NUM_MOTOR; j++) {
                    double sum, err;
		    
                    sum = gv->motor_node[j].bias;
                    for (i = 0; i < NUM_SENSE; i++) {
                        sum += WEIGHT_SM(j, i) * k->input[i];
                    }
                    for (i = 0; i < NUM_CONTEXT; i++) {
                        sum += WEIGHT_CM(j, i) * k->context[i];
                    }
                    err = k->output[j] - sigmoid(sum);

                    for (i = 0; i < NUM_SENSE; i++) {
                        delta_weight[j][i] = err * (k->input[i] - RW_BASELINE);
                    }
                    for (i = 0; i < NUM_CONTEXT; i++) {
                        delta_weight[j][NUM_SENSE+i] = err * (k->context[i] - RW_BASELINE);
                    }
                }

                /* Scale deltas if we have an asymptote: */
                if (gv->target_value > 0.0) {
                    for (j = 0; j < NUM_MOTOR; j++) {
                        for (i = 0; i < NUM_SENSE+NUM_CONTEXT; i++) {
                            delta_weight[j][i] = delta_weight[j][i] * (gv->target_value-gv->weight[j][i]) / gv->target_value;
                        }
                    }
                }

                break;
            }
            default: {
            }
        }

        /* Now apply the deltas to existing weights, incorporating weight decay and ensuring */
        /* weights are non-negative: */

        for (j = 0; j < NUM_MOTOR; j++) {
            for (i = 0; i < NUM_SENSE+NUM_CONTEXT; i++) {
                gv->weight[j][i] = MAX(0.0, (1.0-gv->weight_decay) * gv->weight[j][i] + 
                                             gv->learning_rate * delta_weight[j][i]);
            }
	}
    }

    gv->imitative_delay = tmp_id;
    gv->s_clamp = tmp_sc;
    gv->i_clamp = tmp_ic;
    gv->cycle = 0;
}

/******************************************************************************/

void network_weights_train(GlobalVariables *gv, Boolean update)
{
    double lr;
    int ic, sc;
    int e, i, j;

    ic = gv->i_clamp;
    sc = gv->s_clamp;

    /* Record initial association strenghts: */
    for (i = 0; i < (NUM_SENSE+NUM_CONTEXT); i++) {
        for (j = 0; j < NUM_MOTOR; j++) {
            association[j][i].strength[0] = gv->weight[j][i];
	}
    }

    /* Record the real learning rate - we'll vary this by subject */
    lr = gv->learning_rate;

    for (e = 0; e < gv->epochs; e++) {

        if (gv->print_output) {
            fprintf(stdout, "Epoch %d -------------------------------\n", e+1);
        }

        gv->learning_rate = lr + random_uniform(0.0, LR_NOISE);

        network_weights_train_one_epoch(gv, e);

        for (i = 0; i < (NUM_SENSE+NUM_CONTEXT); i++) {
            for (j = 0; j < NUM_MOTOR; j++) {
                association[j][i].strength[e+1] = gv->weight[j][i];
            }
        }

        if (update) {
            update_graphics(gv);
	}
    }

    /* Restore the various parameters / clamp states: */

    gv->learning_rate = lr;
    gv->i_clamp = ic;
    gv->s_clamp = sc;
}

/******************************************************************************/

void network_initialise_activations(GlobalVariables *gv, Boolean update)
{
    int i;

    for (i = 0; i < NUM_IMPERATIVE; i++) {
        gv->imperative_node[i].bias = gv->imperative_bias;
        gv->imperative_node[i].activation[0] = initial_activation(gv->imperative_node[i].maths, gv->imperative_node[i].bias);
    }
    for (i = 0; i < NUM_SENSE; i++) {
        gv->sensory_node[i].bias = gv->sensory_bias;
        gv->sensory_node[i].activation[0] = initial_activation(gv->sensory_node[i].maths, gv->sensory_node[i].bias);
        gv->sensory_node[i].habituated = FALSE;
    }
    for (i = 0; i < NUM_CONTEXT; i++) {
        gv->context_node[i].bias = gv->context_bias;
        gv->context_node[i].activation[0] = initial_activation(gv->context_node[i].maths, gv->context_node[i].bias);
        gv->context_node[i].habituated = FALSE;
    }
    for (i = 0; i < NUM_MOTOR; i++) {
        gv->motor_node[i].bias = gv->response_bias;
        gv->motor_node[i].activation[0] = initial_activation(gv->motor_node[i].maths, gv->motor_node[i].bias);
        gv->motor_node[i].fired = FALSE;
    }
    gv->cycle = 0;
    update_world(gv);
    if (update) {
        update_graphics(gv);
    }
}

/******************************************************************************/

void network_step_cycle(GlobalVariables *gv, Boolean update)
{
    if (gv->cycle < MAX_CYCLES-1) {
        update_world(gv);
        recirculate_activation(gv);
        if (update) {
            update_graphics(gv);
        }
    }
}

void network_step_trial(GlobalVariables *gv, Boolean update)
{
    while ((network_response(gv) < 0) && (gv->cycle < MAX_CYCLES-1)) {
        update_world(gv);
        recirculate_activation(gv);
        if (update) {
            update_graphics(gv);
        }
    }
}

static void shuffle_trial_mapping(int map[NUM_CONDITIONS])
{
    int i, j, temp;

    for (i = 0; i < NUM_CONDITIONS; i++) {
        j = (i + random_int(NUM_CONDITIONS)) % NUM_CONDITIONS;
        temp = map[i];
        map[i] = map[j];
        map[j] = temp;
    }
}

static void initialise_map(int map[NUM_CONDITIONS])
{
    int c;

    for (c = 0; c < NUM_CONDITIONS; c++) {
        map[c] = c;
    }
    shuffle_trial_mapping(map);
    shuffle_trial_mapping(map);
    shuffle_trial_mapping(map);
    shuffle_trial_mapping(map);
}

void network_step_block(GlobalVariables *gv, SubjectResults *result, int block_size, Boolean update)
{
    int b, c;
    int map[NUM_CONDITIONS];

// This version of the model can drop the response threshold until we make an
// error, then it ups it again. This provides a basic account of error, but
// the fit is poor - too many errors for spatially incompatible trials and
// NO errors for spatially compatible trials.

    initialise_map(map);
    gv->actual_response_threshold = gv->default_response_threshold;

    for (b = 0; b < block_size; b++) {
        shuffle_trial_mapping(map);
        shuffle_trial_mapping(map);
        shuffle_trial_mapping(map);
        shuffle_trial_mapping(map);
        for (c = 0; c < NUM_CONDITIONS; c++) {
            gv->s_clamp = gv->sensory_clamp_settings[map[c]];
            gv->i_clamp = gv->imperative_clamp_settings[map[c]];

            network_initialise_activations(gv, update);
            network_step_trial(gv, update);

            result->response[b][map[c]] = network_response(gv);
            result->rt[b][map[c]]       = (gv->cycle - TIME_ZERO);

            if (gv->dynamic_response_threshold) {
                if (network_response(gv) == (gv->i_clamp - 1)) {
                    /* Response is correct - lower the threshold */
                    gv->actual_response_threshold -= 0.01;
                }
                else {
                    /* Response is an error - reset the threshold */
                    gv->actual_response_threshold = gv->default_response_threshold;
                }
            }
	}
    }
    gv->s_clamp = 0;
    gv->i_clamp = 0;
}

/******************************************************************************/

void src_model_initialise(GlobalVariables *gv)
{
    gv->actual_response_threshold = gv->default_response_threshold;
    network_initialise_activations(gv, TRUE);
}

/******************************************************************************/
