#ifndef _mirror_h_

#define _mirror_h_

#include <stdlib.h>
#include "cs_maths.h"

// This is the cycle at which the imperative stimulus appears. Set it to, say,
// 250, when varying SOA, or perhaps even more to ensure the net settles to a
// stable "prepared" state before stimulus presentation

#define TIME_ZERO 500

typedef enum boolean {
    FALSE, TRUE
} Boolean;

typedef enum learning_type {
    LR_HEBB, LR_QEBB, LR_RW_A, LR_RW_B, NUM_LR
} LearningType;

typedef enum training_type {
    TR_CONTINGENT, TR_NONCONTINGENT, TR_SIGNALLED, TR_OPEN_CLOSE, TR_CLOSE_OPEN, TR_OPEN, TR_CLOSE, NUM_TR
} TrainingType;

typedef enum selection_type {
    ST_THRESHOLD, ST_GAP, ST_PROBABILISTIC
} SelectionType;

#define MAX_CYCLES 2500
#define MAX_BLOCKS 100
#define MAX_SUBJECTS 50
#define NUM_SENSE 2
#define NUM_IMPERATIVE 2
#define NUM_CONTEXT 4
#define NUM_MOTOR 2
#define NUM_CONDITIONS 2
#define TRAIN_MAX 1000

#define WEIGHT_SM(J,I) (gv->weight[J][I])
#define WEIGHT_CM(J,I) (gv->weight[J][NUM_SENSE+I])
#define WEIGHT_IM(J,I) (gv->weight[J][NUM_SENSE+NUM_CONTEXT+I])

typedef struct pattern_list {
    double                imperative[NUM_IMPERATIVE];
    double                input[NUM_SENSE];
    double                context[NUM_CONTEXT];
    double                output[NUM_MOTOR];
    struct pattern_list  *next;
} PatternList;

typedef struct node {
    double activation[MAX_CYCLES];
    double bias;
    Boolean clamped;
    Boolean habituated;
    Boolean fired;
    int colour;
    MathsParameters *maths;
} Node;

typedef struct globals_variables {
    double sensory_bias;
    double context_bias;
    double imperative_bias;
    double response_bias;
    double sensory_li;
    double response_li;
    double weight[NUM_MOTOR][NUM_SENSE+NUM_CONTEXT+NUM_IMPERATIVE];
    double strength_sm;
    double strength_im;
    double strength_cm;
    MathsParameters d_maths;
    MathsParameters s_maths;
    MathsParameters r_maths;
    MathsParameters c_maths;
    double sensory_input_strength;
    double imperative_input_strength;
    double context_input_strength[NUM_CONTEXT];
    double subject_im_sd;
    LearningType learning_rule;
    SelectionType selection;
    double learning_rate;
    double weight_decay;
    double target_value;
    TrainingType training_type;
    PatternList *training_data;
    Boolean sensory_habituation;
    Boolean context_habituation;
    double habituation_threshold;
    Boolean dynamic_response_threshold;
    double default_response_threshold;
    double actual_response_threshold;
    Boolean print_output;
    int epochs;
    int s_clamp;
    int i_clamp;
    int soa;
    int cycle;
    int blocks1;
    int blocks2;
    int subjects1;
    int subjects2;
    int subject_count1;
    int subject_count2;

    int imitative_delay;
    Node imperative_node[NUM_IMPERATIVE];
    Node sensory_node[NUM_SENSE];
    Node context_node[NUM_CONTEXT];
    Node motor_node[NUM_MOTOR];

    void *window;
    void *canvas_network;
    void *canvas_act_profiles;
    void *canvas_weight_profiles;
    void *canvas1;
    void *canvas2;
    double c_and_h_scale;

    int imperative_clamp_settings[NUM_CONDITIONS];
    int sensory_clamp_settings[NUM_CONDITIONS];

} GlobalVariables;

typedef struct subject_results {
    int    response[MAX_BLOCKS][NUM_CONDITIONS];
    double rt[MAX_BLOCKS][NUM_CONDITIONS];
} SubjectResults;

typedef struct assoc {
    int colour;    
    double strength[TRAIN_MAX];
} Assoc;

extern Assoc association[NUM_MOTOR][NUM_SENSE+NUM_CONTEXT];

extern char *lr_name[NUM_LR];

extern void update_graphics(GlobalVariables *gv);

extern GlobalVariables *globals_create();
extern void globals_destroy(GlobalVariables *globals);

extern void network_set_training_data(GlobalVariables *gv, TrainingType tt);

extern void network_weights_initialise(GlobalVariables *gv);
extern void network_weights_train(GlobalVariables *gv, Boolean update);

extern void network_initialise_activations(GlobalVariables *gv, Boolean update);
extern void network_step_cycle(GlobalVariables *gv, Boolean update);
extern void network_step_trial(GlobalVariables *gv, Boolean update);
extern void network_step_block(GlobalVariables *gv, SubjectResults *results, int block_size, Boolean update);

#endif
